/* 
   Unix SMB/CIFS implementation.
   SMB client library implementation (Old interface compatibility)
   Copyright (C) Andrew Tridgell 1998
   Copyright (C) Richard Sharpe 2000
   Copyright (C) John Terpstra 2000
   Copyright (C) Tom Jansen (Ninja ISD) 2002 
   Copyright (C) Derrell Lipman 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "includes.h"

#include "include/libsmb_internal.h"

struct smbc_compat_fdlist {
	SMBCFILE * file;
	int fd;
	struct smbc_compat_fdlist *next, *prev;
};

static SMBCCTX * statcont = NULL;
static int smbc_compat_initialized = 0;
static int smbc_compat_nextfd = 0;
static struct smbc_compat_fdlist * smbc_compat_fd_in_use = NULL;
static struct smbc_compat_fdlist * smbc_compat_fd_avail = NULL;

static pthread_mutex_t smbc_mutex;

#define smbc_enter_mutex()	pthread_mutex_lock(&smbc_mutex)
#define smbc_leave_mutex()	pthread_mutex_unlock(&smbc_mutex)

/* Find an fd and return the SMBCFILE * or NULL on failure */
static SMBCFILE * find_fd(int fd)
{
	struct smbc_compat_fdlist * f = smbc_compat_fd_in_use;
	while (f) {
		if (f->fd == fd) 
			return f->file;
		f = f->next;
	}
	return NULL;
}

/* Add an fd, returns 0 on success, -1 on error with errno set */
static int add_fd(SMBCFILE * file)
{
        struct smbc_compat_fdlist * f = smbc_compat_fd_avail;

	if (f) {
                /* We found one that's available */
                DLIST_REMOVE(smbc_compat_fd_avail, f);

	} else {
                /*
                 * None were available, so allocate one.  Keep the number of
                 * file descriptors determinate.  This allows the application
                 * to allocate bitmaps or mapping of file descriptors based on
                 * a known maximum number of file descriptors that will ever
                 * be returned.
                 */
                if (smbc_compat_nextfd >= FD_SETSIZE) {
                        errno = EMFILE;
                        return -1;
                }

                f = SMB_MALLOC_P(struct smbc_compat_fdlist);
                if (!f) {
                        errno = ENOMEM;
                        return -1;
                }
	
                f->fd = SMBC_BASE_FD + smbc_compat_nextfd++;
        }

	f->file = file;
	DLIST_ADD(smbc_compat_fd_in_use, f);

	return f->fd;
}



/* Delete an fd, returns 0 on success */
static int del_fd(int fd)
{
	struct smbc_compat_fdlist * f = smbc_compat_fd_in_use;

	while (f) {
		if (f->fd == fd) 
			break;
		f = f->next;
	}

	if (f) {
		/* found */
		DLIST_REMOVE(smbc_compat_fd_in_use, f);
                f->file = NULL;
                DLIST_ADD(smbc_compat_fd_avail, f);
		return 0;
	}
	return 1;
}
 


int smbc_init(smbc_get_auth_data_fn fn, int debug)
{
        smbc_enter_mutex();
	if (!smbc_compat_initialized) {
		statcont = smbc_new_context();
		if (!statcont) {
		        smbc_leave_mutex();
			return -1;
		}

		statcont->debug = debug;
		statcont->callbacks.auth_fn = fn;
		
		pthread_mutex_init(&smbc_mutex, NULL);
		
		if (!smbc_init_context(statcont)) {
			smbc_free_context(statcont, False);
			smbc_leave_mutex();
			return -1;
		}

		smbc_compat_initialized = 1;
	}
	smbc_leave_mutex();
	return 0;
}

int smbc_reinit(void)
{
        if (!smbc_compat_initialized)
	        return -1;

	smbc_enter_mutex();

        load_interfaces();  /* Load the list of interfaces ... */

	smbc_leave_mutex();

        return 0;
}

void smbc_remove_unused_servers(unsigned long msec)
{
        smbc_enter_mutex();
	if (statcont)
	    statcont->callbacks.purge_cached_fn(statcont, msec);
	lp_TALLOC_FREE();
	smbc_leave_mutex();
}

SMBCCTX *smbc_set_context(SMBCCTX * context)
{
        SMBCCTX *old_context = statcont;

        if (context) {
                /* Save provided context.  It must have been initialized! */
                statcont = context;

                /* You'd better know what you're doing.  We won't help you. */
		smbc_compat_initialized = 1;
        }
        
        return old_context;
}


int smbc_open(const char *furl, int flags, mode_t mode)
{
	SMBCFILE * file;
	int fd;

	smbc_enter_mutex();
	file = statcont->open(statcont, furl, flags, mode);
	if (!file) {
	        smbc_leave_mutex();
		return -1;
	}

	fd = add_fd(file);
	if (fd == -1) 
		statcont->close_fn(statcont, file);
	smbc_leave_mutex();
	return fd;
}


int smbc_creat(const char *furl, mode_t mode)
{
	SMBCFILE * file;
	int fd;

	smbc_enter_mutex();
	file = statcont->creat(statcont, furl, mode);
	if (!file) {
	        smbc_leave_mutex();
		return -1;
	}

	fd = add_fd(file);
	if (fd == -1) {
		/* Hmm... should we delete the file too ? I guess we could try */
		statcont->close_fn(statcont, file);
		statcont->unlink(statcont, furl);
	}
	smbc_leave_mutex();
	return fd;
}


ssize_t smbc_read(int fd, void *buf, size_t bufsize)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	bufsize = statcont->read(statcont, file, buf, bufsize);
	smbc_leave_mutex();
	return bufsize;
}

ssize_t smbc_write(int fd, void *buf, size_t bufsize)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	bufsize = statcont->write(statcont, file, buf, bufsize);
	smbc_leave_mutex();
	return bufsize;
}

long long smbc_lseek(int fd, long long offset, int whence)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	offset = statcont->lseek(statcont, file, offset, whence);
	smbc_leave_mutex();
	return offset;
}

int smbc_close(int fd)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	del_fd(fd);
	fd = statcont->close_fn(statcont, file);
	smbc_leave_mutex();
	return fd;
}

int smbc_lock(int fd, unsigned long offset, unsigned long len, int timeout, int lock_type)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	lock_type = statcont->lock(statcont, file, offset, len, timeout, lock_type);
	smbc_leave_mutex();
	return lock_type;
}

int smbc_unlock(int fd, unsigned long offset, unsigned long len)
{
        int ret;
        SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	ret = statcont->unlock(statcont, file, offset, len);
	smbc_leave_mutex();
	return ret;
}

int smbc_unlink(const char *fname)
{
        int ret;
        smbc_enter_mutex();
        ret = statcont->unlink(statcont, fname);
	smbc_leave_mutex();
	return ret;
}

int smbc_rename(const char *ourl, const char *nurl)
{
        int ret;
        smbc_enter_mutex();
	ret = statcont->rename(statcont, ourl, statcont, nurl);
	smbc_leave_mutex();
	return ret;
}

int smbc_opendir(const char *durl)
{
	SMBCFILE * file;
	int fd;

	smbc_enter_mutex();
	file = statcont->opendir(statcont, durl);
	if (!file) {
	        smbc_leave_mutex();
		return -1;
	}

	fd = add_fd(file);
	if (fd == -1) 
		statcont->closedir(statcont, file);

	smbc_leave_mutex();
	return fd;
}

int smbc_closedir(int dh) 
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(dh);
	del_fd(dh);
	dh = statcont->closedir(statcont, file);
	smbc_leave_mutex();
	return dh;
}

int smbc_getdents(unsigned int dh, struct smbc_dirent *dirp, int count)
{
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(dh);
	count = statcont->getdents(statcont, file,dirp, count);
	smbc_leave_mutex();
	return count;
}

struct smbc_dirent* smbc_readdir(unsigned int dh)
{
        struct smbc_dirent *d;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(dh);
	d = statcont->readdir(statcont, file);
	smbc_leave_mutex();
	return d;
}

long long smbc_telldir(int dh)
{
        long long off;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(dh);
	off = statcont->telldir(statcont, file);
	smbc_leave_mutex();
	return off;
}

int smbc_lseekdir(int fd, long long offset)
{	
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	ret = statcont->lseekdir(statcont, file, offset);
	smbc_leave_mutex();
	return ret;
}

int smbc_mkdir(const char *durl, mode_t mode)
{
        int ret;
        smbc_enter_mutex();
	ret = statcont->mkdir(statcont, durl, mode);
	smbc_leave_mutex();
	return ret;
}

int smbc_rmdir(const char *durl)
{
        int ret;
        smbc_enter_mutex();
	ret = statcont->rmdir(statcont, durl);
	smbc_leave_mutex();
	return ret;
}

int smbc_stat(const char *url, struct stat *st)
{
        int ret;
        smbc_enter_mutex();
	ret = statcont->stat(statcont, url, st);
	smbc_leave_mutex();
	return ret;
}

int smbc_fstat(int fd, struct stat *st)
{
        int ret;
	SMBCFILE * file;
        smbc_enter_mutex();
	file = find_fd(fd);
	ret = statcont->fstat(statcont, file, st);
	smbc_leave_mutex();
	return ret;
}

int smbc_chmod(const char *url, mode_t mode)
{
        int ret;
	smbc_enter_mutex();
	ret = statcont->chmod(statcont, url, mode);
	smbc_leave_mutex();
	return ret;
}

int smbc_utimes(const char *fname, struct timeval *tbuf)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->utimes(statcont, fname, tbuf);
	smbc_leave_mutex();
	return ret;
}

#ifdef HAVE_UTIME_H
int smbc_utime(const char *fname, struct utimbuf *utbuf)
{
        int ret;
        struct timeval tv[2];

	smbc_enter_mutex();
        if (utbuf == NULL)
                ret = statcont->utimes(statcont, fname, NULL);
	else {
	        tv[0].tv_sec = utbuf->actime;
		tv[1].tv_sec = utbuf->modtime;
		tv[0].tv_usec = tv[1].tv_usec = 0;
		ret = statcont->utimes(statcont, fname, tv);
	}
	smbc_leave_mutex();
	return ret;
}
#endif

int smbc_setxattr(const char *fname,
                  const char *name,
                  const void *value,
                  size_t size,
                  int flags)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->setxattr(statcont, fname, name, value, size, flags);
	smbc_leave_mutex();
	return ret;
}

int smbc_lsetxattr(const char *fname,
                   const char *name,
                   const void *value,
                   size_t size,
                   int flags)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->setxattr(statcont, fname, name, value, size, flags);
	smbc_leave_mutex();
	return ret;
}

int smbc_fsetxattr(int fd,
                   const char *name,
                   const void *value,
                   size_t size,
                   int flags)
{
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	if (file == NULL) {
		errno = EBADF;
		smbc_leave_mutex();
		return -1;
	}
        ret = statcont->setxattr(statcont, file->fname,
				 name, value, size, flags);
	smbc_leave_mutex();
	return ret;
}

int smbc_getxattr(const char *fname,
                  const char *name,
                  const void *value,
                  size_t size)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->getxattr(statcont, fname, name, value, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_lgetxattr(const char *fname,
                   const char *name,
                   const void *value,
                   size_t size)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->getxattr(statcont, fname, name, value, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_fgetxattr(int fd,
                   const char *name,
                   const void *value,
                   size_t size)
{
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	if (file == NULL) {
		errno = EBADF;
		smbc_leave_mutex();
		return -1;
	}
        ret = statcont->getxattr(statcont, file->fname, name, value, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_removexattr(const char *fname,
                     const char *name)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->removexattr(statcont, fname, name);
	smbc_leave_mutex();
	return ret;
}

int smbc_lremovexattr(const char *fname,
                      const char *name)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->removexattr(statcont, fname, name);
	smbc_leave_mutex();
	return ret;
}

int smbc_fremovexattr(int fd,
                      const char *name)
{
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	if (file == NULL) {
		errno = EBADF;
		smbc_leave_mutex();
		return -1;
	}
        ret = statcont->removexattr(statcont, file->fname, name);
	smbc_leave_mutex();
	return ret;
}

int smbc_listxattr(const char *fname,
                   char *list,
                   size_t size)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->listxattr(statcont, fname, list, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_llistxattr(const char *fname,
                    char *list,
                    size_t size)
{
        int ret;
	smbc_enter_mutex();
        ret = statcont->listxattr(statcont, fname, list, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_flistxattr(int fd,
                    char *list,
                    size_t size)
{
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = find_fd(fd);
	if (file == NULL) {
		errno = EBADF;
		smbc_leave_mutex();
		return -1;
	}
        ret = statcont->listxattr(statcont, file->fname, list, size);
	smbc_leave_mutex();
	return ret;
}

int smbc_print_file(const char *fname, const char *printq)
{
        int ret;
	smbc_enter_mutex();
	ret = statcont->print_file(statcont, fname, statcont, printq);
	smbc_leave_mutex();
	return ret;
}

int smbc_open_print_job(const char *fname)
{
        int ret;
	SMBCFILE * file;
	smbc_enter_mutex();
	file = statcont->open_print_job(statcont, fname);
	if (!file) {
	        smbc_leave_mutex();
	        return -1;
	}
	ret = file->cli_fd;
	smbc_leave_mutex();
	return ret;
}

int smbc_list_print_jobs(const char *purl, smbc_list_print_job_fn fn)
{
        int ret;
	smbc_enter_mutex();
	ret = statcont->list_print_jobs(statcont, purl, fn);
	smbc_leave_mutex();
	return ret;
}

int smbc_unlink_print_job(const char *purl, int id)
{
        int ret;
	smbc_enter_mutex();
	ret = statcont->unlink_print_job(statcont, purl, id);
	smbc_leave_mutex();
	return ret;
}

static int _name_query_fd = -1;

static inline int
ensure_name_query_socket()
{
    if(_name_query_fd >= 0)
        return 1;
    
    _name_query_fd = open_socket_in(SOCK_DGRAM, 0, 3,
                                    interpret_addr(lp_socket_address()), 1);

    if(_name_query_fd < 0) {
        printf("can't open name query socket\n");
        return 0;
    }

    set_socket_options(_name_query_fd, "SO_BROADCAST");

    return 1;
}

int nmbc_reinit()
{
    if(_name_query_fd >= 0) {
        close(_name_query_fd);
        _name_query_fd = -1;
    }
    
    return 0;
}

struct in_addr *nmbc_name_query(const char *name, int name_type, BOOL bcast,
                                BOOL recursive, struct in_addr to_ip,
                                int *count, int *flags, BOOL *timed_out)
{
    struct in_addr *ret = NULL;
    smbc_enter_mutex();
    if(ensure_name_query_socket()) {
        ret = name_query(_name_query_fd, name, name_type, bcast, recursive,
                         to_ip, count, flags, timed_out);
    }
    smbc_leave_mutex();
    return ret;
}
