
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/modversions.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/fs.h>

#include "smbredir.h"
#include "../smbremote/smbc_api.h"

MODULE_LICENSE    ("GPL");
MODULE_AUTHOR     ("Jonathan Daub / NETGEAR Inc.");
MODULE_DESCRIPTION("smbredir");
EXPORT_NO_SYMBOLS;

static wait_queue_head_t g_wait_queue;
static char              g_running = 0;
static SMBC_MSG *        g_pending_msg = NULL;
static SMBC_ACK_INFO     g_pending_ack;

int smbredir_open(struct inode *inode, struct file *filp);
int smbredir_release(struct inode *inode, struct file *filp);
ssize_t smbredir_read(struct file *filp, char *buf, size_t count, loff_t *offset);
ssize_t smbredir_write(struct file *filp, const char *buf, size_t count, loff_t *offset);
loff_t smbredir_llseek(struct file *filp, loff_t offset, int whence);
int smbredir_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);

static struct file_operations g_fops = {
    NULL,                	// struct module *owner;
    smbredir_llseek,		// loff_t (*llseek) (struct file *, loff_t, int);
    smbredir_read,		// ssize_t (*read) (struct file *, char *, size_t, loff_t *);
    smbredir_write,		// ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
    NULL,                       // int (*readdir) (struct file *, void *, filldir_t);
    NULL,                       // unsigned int (*poll) (struct file *, struct poll_table_struct *);
    smbredir_ioctl,	        // int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
    NULL,                       // int (*mmap) (struct file *, struct vm_area_struct *);
    smbredir_open,	        // int (*open) (struct inode *, struct file *);
    NULL,                       // int (*flush) (struct file *);
    smbredir_release,	        // int (*release) (struct inode *, struct file *);
    NULL, 			// int (*fsync) (struct file *, struct dentry *, int datasync);
    NULL, 			// int (*fasync) (int, struct file *, int);
    NULL, 			// int (*lock) (struct file *, int, struct file_lock *);
    NULL, 			// ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
    NULL, 		      	// ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
    NULL, 		     	// ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
    NULL 	 	 	// unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
};

int smbredir_open(struct inode *inode, struct file *filp)
{
    SMBC_OPEN_INFO open_info;
    SMBC_ACK_INFO ack_info;

    //printk("smbredir_open %x, mode %x, flags %x, filp %p\n", inode->i_rdev, filp->f_mode, filp->f_flags, filp);
    if (0 == MINOR(inode->i_rdev))
	return 0;
    if (1 != MINOR(inode->i_rdev))
	return -1;

    open_info.fname    = "smb://:@jdaub-coe/media/sigma.hds";
    open_info.flags    = filp->f_flags;
    open_info.mode     = filp->f_mode;
    open_info.hdr.type = SMBC_MSG_OPEN;
    open_info.hdr.size = sizeof(SMBC_OPEN_INFO);
    if (smbredir_send_recv_msg(&open_info, &ack_info) < 0)
	return -1;
    if (ack_info.value.iValue < 0)
	return -1;
    filp->private_data = (void *)ack_info.value.iValue;

    //printk("smbredir_open ->%d\n", ack_info.value.iValue);
    return 0;
}

int smbredir_release(struct inode *inode, struct file *filp)
{
    SMBC_CLOSE_INFO close_info;
    SMBC_ACK_INFO ack_info;

    //printk("smbredir_release, filp %x\n", filp);

    if (!filp->private_data)
	return -1;

    close_info.fd       = (int)filp->private_data;
    close_info.hdr.type = SMBC_MSG_CLOSE;
    close_info.hdr.size = sizeof(SMBC_CLOSE_INFO);
    if (smbredir_send_recv_msg(&close_info, &ack_info) < 0)
	return -1;

    filp->private_data = NULL;
    return 0;
}

ssize_t smbredir_read(struct file *filp, char *buf, size_t count, loff_t *offset)
{
    SMBC_READ_INFO read_info;
    SMBC_ACK_INFO ack_info;

    //printk("smbredir_read, filp %x, count %d\n", filp, count);

    if (!filp->private_data)
	return -1;

    read_info.fd       = (int)filp->private_data;
    read_info.count    = count;
    read_info.buf      = buf;
    read_info.hdr.type = SMBC_MSG_READ;
    read_info.hdr.size = sizeof(SMBC_READ_INFO);
    if (smbredir_send_recv_msg(&read_info, &ack_info) < 0)
	return -1;
    if (ack_info.value.iValue > 0)
	(*offset) += ack_info.value.iValue;
    return ack_info.value.iValue;
}

ssize_t smbredir_write(struct file *filp, const char *buf, size_t count, loff_t *offset)
{
    SMBC_WRITE_INFO write_info;
    SMBC_ACK_INFO ack_info;

    //printk("smbredir_write %d\n", count);

    if (!filp->private_data)
	return -1;

    write_info.fd       = (int)filp->private_data;
    write_info.buf      = (char *)buf;
    write_info.count    = count;
    write_info.hdr.type = SMBC_MSG_WRITE;
    write_info.hdr.size = sizeof(SMBC_WRITE_INFO);
    if (smbredir_send_recv_msg(&write_info, &ack_info) < 0)
	return -1;
    if (ack_info.value.iValue > 0)
	(*offset) += ack_info.value.iValue;
    return ack_info.value.iValue;
}

loff_t smbredir_llseek(struct file *filp, loff_t offset, int whence)
{
    SMBC_LSEEK_INFO lseek_info;
    SMBC_ACK_INFO ack_info;

    //printk("smbredir_llseek %lld whence %d\n", offset, whence);

    if (!filp->private_data)
	return -1;

    lseek_info.fd       = (int)filp->private_data;
    lseek_info.offset   = offset;
    lseek_info.whence   = whence;
    lseek_info.hdr.type = SMBC_MSG_LSEEK;
    lseek_info.hdr.size = sizeof(SMBC_LSEEK_INFO);
    if (smbredir_send_recv_msg(&lseek_info, &ack_info) < 0)
	return -1;
    return ack_info.value.llValue;
}

int smbredir_send_recv_msg(SMBC_MSG *msg, SMBC_ACK_INFO *ack_msg)
{
    wait_queue_t wait;
    int retval;

    //printk("smbredir_send_recv_msg\n");

    if (g_pending_msg) {
	printk("smbredir: Busy!\n");
	return -1;
    }

    g_pending_msg = msg;
    g_pending_ack.hdr.size = 0;
    wake_up_interruptible(&g_wait_queue);

    while (!g_pending_ack.hdr.size) {

	if (!g_running)
	    return -1;

	init_waitqueue_entry(&wait, current);
	add_wait_queue(&g_wait_queue, &wait);
	current->state = TASK_INTERRUPTIBLE;

	retval = 1;
	if (!g_pending_ack.hdr.size) {
	    //printk("smbredir_send_recv_msg: Sleeping\n");
	    retval = schedule_timeout(HZ * 3);
	    //printk("smbredir_send_recv_msg: Waking up %d\n", retval);
	}

	current->state = TASK_RUNNING;
	remove_wait_queue(&g_wait_queue, &wait);
    }

    memcpy(ack_msg, &g_pending_ack, g_pending_ack.hdr.size);

    //printk("smbredir_send_recv_msg -> 0\n");
    return 0;
}

int smbredir_wait_msg(SMBC_MSG *msg)
{
    wait_queue_t wait;
    int retval;

    //printk("smbredir_wait_msg\n");
    while (!g_pending_msg) {

	if (!g_running)
	    return -1;

	init_waitqueue_entry(&wait, current);
	add_wait_queue(&g_wait_queue, &wait);
	current->state = TASK_INTERRUPTIBLE;

	retval = 1;
	if (!g_pending_msg) {
	    //printk("smbredir_wait_msg: Sleeping\n");
	    retval = schedule_timeout(HZ * 60);
	    //printk("smbredir_wait_msg: Waking up %d\n", retval);
	}

	current->state = TASK_RUNNING;
	remove_wait_queue(&g_wait_queue, &wait);
    }

    memcpy(msg, g_pending_msg, g_pending_msg->size);
    g_pending_msg = NULL;
    //printk("smbredir_wait_msg -> 0\n");
    return 0;
}

int smbredir_reply_msg(SMBC_MSG *msg)
{
    //printk("smbredir_reply_msg\n");
    if (msg->type != SMBC_MSG_ACK || msg->size > sizeof(g_pending_ack))
	return -1;
    memcpy(&g_pending_ack, msg, msg->size);
    wake_up_interruptible(&g_wait_queue);
    return 0;
}

int smbredir_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
{
    switch (ioctl_num) {
    case SMBREDIR_IOCTL_GET_VERSION:
	return 0x0100;
    case SMBREDIR_IOCTL_WAIT_MSG:
	return smbredir_wait_msg((SMBC_MSG *)ioctl_param);
    case SMBREDIR_IOCTL_REPLY_MSG:
	return smbredir_reply_msg((SMBC_MSG *)ioctl_param);
    default:
	printk("smbredir_ioctl: Unknown ioctl %d\n", ioctl_num);
	return -1;
    }
}

int init_module()
{
    int ret_val;

    init_waitqueue_head(&g_wait_queue);

    ret_val = register_chrdev(SMBREDIR_MAJOR_NUM, SMBREDIR_DEVICE_NAME, &g_fops);
    if (ret_val < 0) {
	printk ("smbredir registration failed, error %d\n", ret_val);
	return ret_val;
    }
    printk("smbredir device loaded.\n");
    
    g_running = 1;
    return ret_val;
}

void cleanup_module()
{
    g_running = 0;
    wake_up_interruptible(&g_wait_queue);
    unregister_chrdev(SMBREDIR_MAJOR_NUM, SMBREDIR_DEVICE_NAME);
}


