
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <libsmbclient.h>

#include "smbc_api.h"

#define smbc_enter_mutex()	pthread_mutex_lock(&smbc_mutex)
#define smbc_leave_mutex()	pthread_mutex_unlock(&smbc_mutex)
#define FIELD_OFFSET(type, field)  ((long)&(((type *)0)->field))

static int smbc_fd = -1;
static int smbc_initialized = 0;
static pthread_mutex_t smbc_mutex;
static smbc_get_auth_data_fn smbc_auth_fn = NULL;

int smbc_init(smbc_get_auth_data_fn fn, int debug)
{
    struct sockaddr_in addr;

    if (!smbc_initialized) {
	pthread_mutex_init(&smbc_mutex, NULL);
	smbc_initialized = 1;
    }

    smbc_enter_mutex();
    if (smbc_fd < 0) {
	memset(&addr, 0, sizeof(addr));
	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	addr.sin_port        = htons(SMBC_REMOTE_PORT);
	addr.sin_family      = AF_INET;
	smbc_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (smbc_fd >= 0 && connect(smbc_fd, &addr, sizeof(addr)) < 0) {
	    close(smbc_fd);
	    smbc_fd = -1;
	}
    }
    smbc_leave_mutex();
    
    return (smbc_fd >= 0);
}

int smbc_reinit(void)
{
    int ret;
    SMBC_MSG msg;
    SMBC_ACK_INFO ack_info;

    msg.type = SMBC_MSG_REINIT;
    msg.size = sizeof(SMBC_MSG);
    ret = smbc_send_recv_msg(&msg, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_open(const char *fname, int flags, mode_t mode)
{
    int ret;
    SMBC_OPEN_INFO open_info;
    SMBC_ACK_INFO ack_info;

    open_info.fname    = fname;
    open_info.flags    = flags;
    open_info.mode     = mode;
    open_info.hdr.type = SMBC_MSG_OPEN;
    open_info.hdr.size = sizeof(SMBC_OPEN_INFO);
    ret = smbc_send_recv_msg(&open_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_creat(const char *furl, mode_t mode)
{
    return smbc_open(furl, O_CREAT|O_WRONLY|O_TRUNC, mode);
}

ssize_t smbc_read(int fd, void *buf, size_t count)
{
    int ret;
    SMBC_READ_INFO read_info;
    SMBC_ACK_INFO ack_info;

    read_info.fd       = fd;
    read_info.count    = count;
    read_info.buf      = buf;
    read_info.hdr.type = SMBC_MSG_READ;
    read_info.hdr.size = sizeof(SMBC_READ_INFO);
    ret = smbc_send_recv_msg(&read_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

ssize_t smbc_write(int fd, void *buf, size_t count)
{
    int ret;
    SMBC_WRITE_INFO write_info;
    SMBC_ACK_INFO ack_info;

    write_info.fd       = fd;
    write_info.buf      = buf;
    write_info.count    = count;
    write_info.hdr.type = SMBC_MSG_WRITE;
    write_info.hdr.size = sizeof(SMBC_WRITE_INFO);
    ret = smbc_send_recv_msg(&write_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

long long smbc_lseek(int fd, long long offset, int whence)
{
    long long ret;
    SMBC_LSEEK_INFO lseek_info;
    SMBC_ACK_INFO ack_info;

    lseek_info.fd       = fd;
    lseek_info.offset   = offset;
    lseek_info.whence   = whence;
    lseek_info.hdr.type = SMBC_MSG_LSEEK;
    lseek_info.hdr.size = sizeof(SMBC_LSEEK_INFO);
    ret = smbc_send_recv_msg(&lseek_info, &ack_info) ? ack_info.value.llValue : -1;
    return ret;
}

int smbc_close(int fd)
{
    int ret;
    SMBC_CLOSE_INFO close_info;
    SMBC_ACK_INFO ack_info;

    close_info.fd       = fd;
    close_info.hdr.type = SMBC_MSG_CLOSE;
    close_info.hdr.size = sizeof(SMBC_CLOSE_INFO);
    ret = smbc_send_recv_msg(&close_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_unlink(const char *furl)
{
    int ret;
    SMBC_UNLINK_INFO unlink_info;
    SMBC_ACK_INFO ack_info;

    unlink_info.furl     = furl;
    unlink_info.hdr.type = SMBC_MSG_UNLINK;
    unlink_info.hdr.size = sizeof(SMBC_UNLINK_INFO);
    ret = smbc_send_recv_msg(&unlink_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_lock(int fd, unsigned long offset, unsigned long len, int timeout, int lock_type)
{
    int ret;
    SMBC_LOCK_INFO lock_info;
    SMBC_ACK_INFO ack_info;

    lock_info.fd        = fd;
    lock_info.offset    = offset;
    lock_info.len       = len;
    lock_info.timeout   = timeout;
    lock_info.lock_type = lock_type;
    lock_info.hdr.type  = SMBC_MSG_LOCK;
    lock_info.hdr.size  = sizeof(SMBC_LOCK_INFO);
    ret = smbc_send_recv_msg(&lock_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_unlock(int fd, unsigned long offset, unsigned long len)
{
    int ret;
    SMBC_UNLOCK_INFO unlock_info;
    SMBC_ACK_INFO ack_info;

    unlock_info.fd        = fd;
    unlock_info.offset    = offset;
    unlock_info.len       = len;
    unlock_info.hdr.type  = SMBC_MSG_UNLOCK;
    unlock_info.hdr.size  = sizeof(SMBC_UNLOCK_INFO);
    ret = smbc_send_recv_msg(&unlock_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_rename(const char *ourl, const char *nurl)
{
    int ret;
    SMBC_RENAME_INFO rename_info;
    SMBC_ACK_INFO ack_info;

    rename_info.ourl     = ourl;
    rename_info.nurl     = nurl;
    rename_info.hdr.type = SMBC_MSG_RENAME;
    rename_info.hdr.size = sizeof(SMBC_RENAME_INFO);
    ret = smbc_send_recv_msg(&rename_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_opendir(const char *fname)
{
    int ret;
    SMBC_OPENDIR_INFO opendir_info;
    SMBC_ACK_INFO ack_info;

    opendir_info.fname    = fname;
    opendir_info.hdr.type = SMBC_MSG_OPENDIR;
    opendir_info.hdr.size = sizeof(SMBC_OPENDIR_INFO);
    ret = smbc_send_recv_msg(&opendir_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_closedir(int fd)
{
    int ret;
    SMBC_CLOSEDIR_INFO closedir_info;
    SMBC_ACK_INFO ack_info;

    closedir_info.fd      = fd;
    closedir_info.hdr.type = SMBC_MSG_CLOSEDIR;
    closedir_info.hdr.size = sizeof(SMBC_CLOSEDIR_INFO);
    ret = smbc_send_recv_msg(&closedir_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

struct smbc_dirent *smbc_readdir(unsigned int fd)
{    
    struct smbc_dirent *ret;
    SMBC_READDIR_INFO readdir_info;
    SMBC_ACK_INFO ack_info;

    readdir_info.fd       = fd;
    readdir_info.hdr.type = SMBC_MSG_READDIR;
    readdir_info.hdr.size = sizeof(SMBC_READDIR_INFO);
    ret = smbc_send_recv_msg(&readdir_info, &ack_info) ? ack_info.value.pvValue : NULL;
    return ret;
}

long long smbc_telldir(int dh)
{
    long long ret;
    SMBC_TELLDIR_INFO telldir_info;
    SMBC_ACK_INFO ack_info;

    telldir_info.dh       = dh;
    telldir_info.hdr.type = SMBC_MSG_TELLDIR;
    telldir_info.hdr.size = sizeof(SMBC_TELLDIR_INFO);
    ret = smbc_send_recv_msg(&telldir_info, &ack_info) ? ack_info.value.llValue : -1;
    return ret;
}

int smbc_lseekdir(int fd, long long offset)
{
    int ret;
    SMBC_LSEEKDIR_INFO lseekdir_info;
    SMBC_ACK_INFO ack_info;

    lseekdir_info.fd       = fd;
    lseekdir_info.offset   = offset;
    lseekdir_info.hdr.type = SMBC_MSG_LSEEKDIR;
    lseekdir_info.hdr.size = sizeof(SMBC_LSEEKDIR_INFO);
    ret = smbc_send_recv_msg(&lseekdir_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_mkdir(const char *fname, mode_t mode)
{
    int ret;
    SMBC_MKDIR_INFO mkdir_info;
    SMBC_ACK_INFO ack_info;

    mkdir_info.fname    = fname;
    mkdir_info.mode     = mode;
    mkdir_info.hdr.type = SMBC_MSG_MKDIR;
    mkdir_info.hdr.size = sizeof(SMBC_MKDIR_INFO);
    ret = smbc_send_recv_msg(&mkdir_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_rmdir(const char *durl)
{
    int ret;
    SMBC_RMDIR_INFO rmdir_info;
    SMBC_ACK_INFO ack_info;

    rmdir_info.durl     = durl;
    rmdir_info.hdr.type = SMBC_MSG_RMDIR;
    rmdir_info.hdr.size = sizeof(SMBC_RMDIR_INFO);
    ret = smbc_send_recv_msg(&rmdir_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_stat(const char *fname, struct stat *st)
{
    int ret;
    SMBC_STAT_INFO stat_info;
    SMBC_ACK_INFO ack_info;

    stat_info.fname     = fname;
    stat_info.st        = st;
    stat_info.hdr.type  = SMBC_MSG_STAT;
    stat_info.hdr.size  = sizeof(SMBC_STAT_INFO);
    ret = smbc_send_recv_msg(&stat_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_fstat(int fd, struct stat *st)
{
    int ret;
    SMBC_FSTAT_INFO fstat_info;
    SMBC_ACK_INFO ack_info;

    fstat_info.fd       = fd;
    fstat_info.st       = st;
    fstat_info.hdr.type = SMBC_MSG_FSTAT;
    fstat_info.hdr.size = sizeof(SMBC_FSTAT_INFO);
    ret = smbc_send_recv_msg(&fstat_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

#if 0
int smbc_chown(const char *url, uid_t owner, gid_t group)
{
    int ret;
    SMBC_CHOWN_INFO chown_info;
    SMBC_ACK_INFO ack_info;

    chown_info.url      = url;
    chown_info.owner    = owner;
    chown_info.group    = group;
    chown_info.hdr.type = SMBC_MSG_CHOWN;
    chown_info.hdr.size = sizeof(SMBC_CHOWN_INFO);
    ret = smbc_send_recv_msg(&chown_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}

int smbc_chmod(const char *url, mode_t mode)
{
    int ret;
    SMBC_CHMOD_INFO chmod_info;
    SMBC_ACK_INFO ack_info;

    chmod_info.url      = url;
    chmod_info.mode     = mode;
    chmod_info.hdr.type = SMBC_MSG_CHMOD;
    chmod_info.hdr.size = sizeof(SMBC_CHMOD_INFO);
    ret = smbc_send_recv_msg(&chmod_info, &ack_info) ? ack_info.value.iValue : -1;
    return ret;
}
#endif

void smbc_remove_unused_servers(unsigned long msec)
{
    SMBC_REMOVEUNUSEDSERVERS_INFO rus_info;
    SMBC_ACK_INFO ack_info;

    rus_info.msec     = msec;
    rus_info.hdr.type = SMBC_MSG_REMOVEUNUSEDSERVERS;
    rus_info.hdr.size = sizeof(SMBC_REMOVEUNUSEDSERVERS_INFO);
    smbc_send_recv_msg(&rus_info, &ack_info);
}

int nmbc_name_query(const char *name, struct in_addr bcast_addr, struct in_addr *ip_addr)
{
    NMBC_NAMEQUERY_INFO nq_info;
    SMBC_ACK_INFO ack_info;
    
    nq_info.name = name;
    memcpy(&nq_info.bcast_addr, &bcast_addr, sizeof(struct in_addr));
    nq_info.ip_addr = ip_addr;
    nq_info.hdr.type = NMBC_MSG_NAMEQUERY;
    nq_info.hdr.size = sizeof(NMBC_NAMEQUERY_INFO);
    return smbc_send_recv_msg(&nq_info, &ack_info) ? ack_info.value.iValue : -1;
}

int smbc_send_recv_msg(SMBC_MSG *msg, SMBC_ACK_INFO *ack_msg)
{
    int ret;

    smbc_enter_mutex();
    ret = smbc_send_msg(msg);
    if (ret)
	ret = smbc_recv_msg(ack_msg, sizeof(SMBC_ACK_INFO));
	if(ack_msg->_errno)
		errno=ack_msg->_errno;
    smbc_leave_mutex();
    return ret;
}

int smbc_send_msg(SMBC_MSG *msg)
{
    int i, j;

    //printf("smbc_send_msg: Sending msg %d\n", msg->type);
    for (i = 0; i < msg->size; i += j) {

	j = send(smbc_fd, ((char *)msg)+i, msg->size - i, 0);
	if (j < 0) {
	    printf("smbc_send_msg: Failed to send %d, error %d\n", j, errno);
	    return 0;
	}
    }
    return 1;
}

int smbc_recv_msg(SMBC_MSG *msg, int size)
{
    int i, j;

    //printf("smbc_recv_msg: Receive msg into %d\n", size);
    for (i = 0; i < sizeof(SMBC_MSG); i += j) {

	j = recv(smbc_fd, ((char *)msg) + i, sizeof(SMBC_MSG) - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_recv_msg: Receive msg %d size %d\n", msg->type, msg->size);

    if (msg->size > size) {
	printf("smbc_recv_msg: Message too large %d\n", msg->size);
	return 0;
    }

    for (i; i < msg->size; i+= j) {

	j = recv(smbc_fd, ((char *)msg) + i, msg->size - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_recv_msg: Read %d bytes\n", i);
    return i;
}

char *smbc_strcpyn(char *dst, char *src, int len)
{
    char *ptr = dst;

    if (!len)
	return 0;

    while (*src) {

	if (len <= 1) {

	    *ptr = 0;
	    return dst;
	}

	*ptr = *src;
	ptr++;
	src++;
	len--;
    }

    *ptr = 0;
    return dst;
}
