#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <malloc.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <libsmbclient.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <linux/kernel.h>

typedef long long LONGLONG, *PLONGLONG;

#define FIELD_OFFSET(type, field)  ((long)&(((type *)0)->field))

typedef struct FIOFUNCSTRUCT {
  int		fh;
  int		(*close)(int);
  ssize_t	(*read)(int, void *, size_t);
  ssize_t       (*write)(int, void *, size_t);
  off_t	        (*seek)(int, off_t, int);
} FIO_FUNC;

#define fio_read(_fio, _buf, _len) (_fio)->read((_fio)->fh, _buf, _len)
#define fio_write(_fio, _buf, _len) (_fio)->write((_fio)->fh, _buf, _len)
#define fio_seek(_fio, _off, _from) (_fio)->seek((_fio)->fh, _off, _from)
#define fio_close(_fio) (_fio)->close((_fio)->fh)

ssize_t fio_smbc_read(int fh, void *buf, size_t len);
ssize_t fio_smbc_write(int fh, void *buf, size_t len);
__off64_t fio_smbc_lseek(int fh, __off64_t off, int from);
int fio_smbc_close(int fh);

void fio_auth_fn(const char *server, const char *share,
		 char *workgroup, int wgmaxlen, char *username, int unmaxlen,
		 char *password, int pwmaxlen)
{
  printf("fio_auth_fn\n");
#if 0
  if (server && server[0])
    printf("  server: %s\n", server);
  if (share && share[0])
    printf("  share: %s\n", share);
  if (workgroup && workgroup[0])
    printf("  workgroup: %s\n", workgroup);
  if (username && username[0])
    printf("  username: %s\n", username);
#endif
  username[0] = password[0] = 0;
}


//-------------------------------------------------------------------------
// String utility functions.
//-------------------------------------------------------------------------

LONGLONG getlongtime(void)
{
    LONGLONG t;
    struct timeb tv;
    ftime(&tv);
    t = ((LONGLONG)tv.time * 1000) + (LONGLONG)tv.millitm;
    return t;
}

int Util_RemoveFileSpec(char *filename)
{
    char *filespec;
    char *next;

    filespec = filename;
    for (next = filespec; *next; next++) {

	if (*next == '/') {
	    filespec = next;
	}
	else if (*next == ':') {
	    if (next[1] == '/')
		next++;
	    filespec = next + 1;
	}
    }

    if (*filespec == 0)
	return 0;

    if ((filespec == filename) && (*filespec == '/')) {

	if (filespec[1]) {

	    filespec[1] = 0;
	    return 1;
	}
	return 0;
    }

    *filespec = 0;
    return 1;
}

int Util_IsUNCPath(char *filename)
{
    if ((filename[0] == '/') && (filename[1] == '/'))
	return 1;
    return 0;
}

char *Util_GetNextComponent(char *filename)
{
    while (*filename) {
	if (*filename == '/') {
	    filename++;
	    if (*filename == '/')
		filename++;
	    return filename;
	}
	filename++;
    }
    return 0;
}

char *strcpyn(char *dst, char *src, int len)
{
    char *ptr = dst;

    if (!len)
	return NULL;

    while (*src) {

	if (len <= 1) {

	    *ptr = 0;
	    return dst;
	}

	*ptr = *src;
	ptr++;
	src++;
	len--;
    }

    *ptr = 0;
    return dst;
}

//-------------------------------------------------------------------------
// File i/o functions
//-------------------------------------------------------------------------
pthread_mutex_t g_smbc_mutex;

int fio_issmbpath(char *path)
{
  return strncasecmp(path, "smb:", 4) ? 0 : 1;
}

int fio_open(char *path, int flags, int mode, FIO_FUNC *fio)
{
#ifdef LINUX
  if (fio_issmbpath(path)) {

      //fio_mount_smb_path(path + 4, 0);
      //return -1;

    fio->close = fio_smbc_close;
    fio->read  = fio_smbc_read;
    fio->write = fio_smbc_write;
    fio->seek  = NULL;
    //fio->seek  = fio_smbc_lseek;

    if (fio_init_smb() < 0)
      return -1;

    //pthread_mutex_lock(&g_smbc_mutex);
    fio->fh = smbc_open(path, flags, mode);
    //pthread_mutex_unlock(&g_smbc_mutex);

    if (fio->fh < 0) 
      return -1;

    //printf("fio_open, fio fh = %d\n", fio->fh);
    return 0;
  }
#endif

  fio->close = close;
  fio->read  = read;
  fio->write = write;
  fio->seek  = lseek;

  fio->fh = open(path, flags, mode);
  if (fio->fh < 0) 
    return -1;

  return 0;
}

int fio_stat(char *path, struct stat *st)
{
#ifdef LINUX
  int r;

  if (fio_issmbpath(path)) {
    
    if (fio_init_smb() < 0)
      return -1;

    pthread_mutex_lock(&g_smbc_mutex);
    r = smbc_stat(path, st);
    pthread_mutex_unlock(&g_smbc_mutex);

    return r;
  }
#endif

  return stat(path, st);
}

ssize_t fio_smbc_read(int fh, void *buf, size_t len)
{
  ssize_t r;

  //pthread_mutex_lock(&g_smbc_mutex);
  r = smbc_read(fh, buf, len);
  //pthread_mutex_unlock(&g_smbc_mutex);

  return r;
}

ssize_t fio_smbc_write(int fh, void *buf, size_t len)
{
  pthread_mutex_lock(&g_smbc_mutex);
  len = smbc_write(fh, buf, len);
  pthread_mutex_unlock(&g_smbc_mutex);

  return len;
}

__off64_t fio_smbc_lseek(int fh, __off64_t off, int from)
{
  //pthread_mutex_lock(&g_smbc_mutex);
  off = smbc_lseek(fh, off, from);
  //pthread_mutex_unlock(&g_smbc_mutex);

  return off;
}

int fio_smbc_close(int fh)
{
  int r;

  //pthread_mutex_lock(&g_smbc_mutex);
  r = smbc_close(fh);
  //pthread_mutex_unlock(&g_smbc_mutex);

  return r;
}

int fio_init_smb(void)
{
  static int initflag = -1;

  pthread_mutex_lock(&g_smbc_mutex);

  if (initflag < 0)
    initflag = smbc_init(fio_auth_fn, 0);

  pthread_mutex_unlock(&g_smbc_mutex);

  return initflag;
}

int fio_mount_smb_path(char *path, char *mntpath, int mntpathsize)
{
    char cmd[1024];
    char dir[1024];
    char *next;
    int status;

    printf("Mounting %s\n", path);

    if (!Util_IsUNCPath(path)) {
	printf("Not a UNC path\n");
	return -1;
    }

    next = Util_GetNextComponent(path + 2);
    if (!next) {
	printf("Invalid path name\n");
	return -1;
    }
    
    do {

	next = Util_GetNextComponent(next);
	if (!next) {
	    printf("Couldn't mount any directories in %s\n", path);
	    return -1;
	}

	next[-1] = 0;
	sprintf(cmd, "smbmount %s /t -o password=\"\"", path);
	printf("Launching %s\n", cmd);

	status = system(cmd);
	if (-1 != status) {
	    printf("smbmount status: %d, exit status %d\n", status, WEXITSTATUS(status));
	    if (0 == WEXITSTATUS(status))
		break;
	}
	else {
	    printf("smbmount failed\n");
	}

	next[-1] = '/';

    } while (1);

    printf("Mounted %s, next = %s\n", path, next);
    return 0;
}

int fio_smbc_makepathto(char *fname)
{
    struct stat st;
    int r;
    char   path[512];

    strcpy(path, fname);
    Util_RemoveFileSpec(path);
    if (!strchr(&path[6], '/'))
	return 0;

    memset(&st, 0, sizeof(st));
    if (smbc_stat(path, &st) >= 0) {

	printf("smbc stat on %s -> mode %x\n", path, st.st_mode);
	return 0;
    }

    printf("smbc stat on %s -> %d\n", path, errno);

    if (fio_smbc_makepathto(path) < 0)
	return -1;

    printf("Creating dir %s\n", path);
    if (smbc_mkdir(path, 0x666) < 0)
	return -1;

    return 0;
}

void *smbtest_test(char *filename)
{
    FIO_FUNC fio;
    int r;
    int i;
    struct sysinfo si;
    long lomem;
    struct stat st;

    fio_init_smb();
    printf("Opening file %s\n", filename);

    lomem = 0x7fffffff;
    i = 0;
    do {

	//smbc_stat(filename, &st);
	r = -1;
	r = fio_open(filename, O_RDWR, 0, &fio);
	if (r < 0) {
	    //printf("Error opening %s (error %d, errno %d)\n", filename, r, errno);
	}
	else {
	    fio_close(&fio);
	}
	
	if (i++ > 100) {
	    smbc_remove_unused_servers(0);
	    i = 0; 
	    sysinfo(&si);
	    //if (si.freeram < lomem) {
		printf("%d bytes free\n", si.freeram);
		lomem = si.freeram;
		//}
	}

	//sleep(1);
    } while (1);
    return 0;
}

void *smbtest_thread(char *filename)
{
  FIO_FUNC fio;
  int len;
  static char buf[65536];
  LONGLONG total, grand_total;
  struct stat st;
  int dh;	
  int locked;
  int r;	
  LONGLONG time, diff;
  __off64_t size, pos;
  struct smbc_dirent *sde;

  printf("THREAD %x STARTING\n", pthread_self());

  fio_init_smb();

//  printf("Stat'ing %s\n", filename);
//  r = fio_stat(filename, &st);
//  if (r < 0) {
//      
//      printf("Couldn't stat %s (error %d, errno %d)\n", filename, r, errno);
//      return 0;
//  }

  if (filename[strlen(filename)-1] == '/') {
//  if (S_ISDIR(st.st_mode)) {

      filename[strlen(filename)-1] = 0;
      printf("Opening directory %s\n", filename);

      dh = smbc_opendir(filename);      
      if (dh < 0) {

	  printf("Couldn't open directory %s (error %d, errno %d)\n", filename, dh, errno);
	  return 0;
      }

      while (sde = smbc_readdir(dh)) 
	  printf("  %s\n", sde->name);

      printf("Sleeping for 20 seconds...\n");
      sleep(20);
      smbc_closedir(dh);
      return 1;
  }

  printf("Opening file %s\n", filename);
  r = fio_open(filename, O_RDWR, 0, &fio);
  if (r < 0) {

    printf("Error opening %s (error %d, errno %d)\n", filename, r, errno);
    return 0;
  }

  if (!fio.seek) {
      size = fio_smbc_lseek(fio.fh, 0, SEEK_END);
      fio_smbc_lseek(fio.fh, 0, SEEK_SET);
  }
  else {
      size = fio.seek(fio.fh, 0, SEEK_END);
      fio.seek(fio.fh, 0, SEEK_SET);
  }
  printf("Total size: %lld\n", size);

  memset(&st, 0, sizeof(st));
  smbc_stat(filename, &st);
  printf("Size from stat: %ld\n", st.st_size);

#if 0
  printf("Locking file for reading...\n");
  r = smbc_lock(fio.fh, 0, 1000, -1, 0);
  while (r < 0) {

      printf("Error locking %s (error %d, errno %d)\n", filename, r, errno);
      //smbc_unlock(fio.fh, 0, 1000);
      sleep(1);
      //return 0;
  }
  printf("locked %s\n", filename);
  locked = 1;

  sleep(10);
  printf("Unlocking %s\n", filename);
  r = smbc_unlock(fio.fh, 0, 1000);
  if (r < 0) 
      printf("Error unlocking %s (error %d, errno %d)\n", filename, r, errno);

  sleep(10);
#endif

#if 1
  total = 0;
  grand_total = 0;
  time = getlongtime();
  do {
    
    len = fio_read(&fio, buf, sizeof(buf));
    if (!len)
	break;
    if (len < 0) {
	printf("Error reading %s (error %d, errno %d)\n", filename, r, errno);
	break;
    }

    //pos = fio.seek(fio.fh, 0, SEEK_CUR);
    //size = fio.seek(fio.fh, 0, SEEK_END);
    //fio.seek(fio.fh, pos, SEEK_SET);
    //printf("Pos: %d  Size: %d\n", pos, size);

    //printf(".");
    total += len;
    grand_total += len;

    if (getlongtime() >= time + 10000) {

	diff = getlongtime() - time;
	printf("[%lld kb] Read %lld kb in %lld msec - %lld kbps\n", grand_total/1024, total/1024, diff, diff ? total/diff : 0);
	time = getlongtime();
	total = 0;
    }

  } while (1);
#endif

  fio_close(&fio);

  if (len < 0)
    printf("THREAD %x ABORTED after %d bytes read, err = %d\n", pthread_self(), total, errno);
  else
    printf("THREAD %x EXITING after %d bytes read (eof), err = %d\n", pthread_self(), total, errno);

  return 0;
}

void *smbtest_write_thread(char *filename)
{
  FIO_FUNC fio;
  int len;
  char buf[65536];
  LONGLONG total, grand_total;
  struct stat st;
  int dh;	
  int locked;
  int r;	
  LONGLONG time, diff;
  __off64_t size, pos;

  printf("WRITE THREAD %x STARTING\n", pthread_self());

  fio_init_smb();

  printf("Creating file %s\n", filename);
  r = fio_open(filename, O_RDWR|O_CREAT|O_TRUNC, 0x666, &fio);
  if (r < 0) {

      fio_smbc_makepathto(filename);

      r = fio_open(filename, O_RDWR|O_CREAT|O_TRUNC, 0x666, &fio);
      if (r < 0) {

	  printf("Error creating %s (error %d, errno %d)\n", filename, r, errno);
	  return 0;
      }
  }

#if 1
  printf("Locking file for writing...\n");
  r = smbc_lock(fio.fh, 0, 1000, -1, 1);
  while (r < 0) {

      printf("Error locking %s (error %d, errno %d)\n", filename, r, errno);
      //smbc_unlock(fio.fh, 0, 1000);
      sleep(1);
      //return 0;
  }
  printf("locked %s\n", filename);
  locked = 1;

  sleep(10);
  printf("Unlocking %s\n", filename);
  r = smbc_unlock(fio.fh, 0, 1000);
  if (r < 0) 
      printf("Error unlocking %s (error %d, errno %d)\n", filename, r, errno);
#endif

  total = 0;
  grand_total = 0;
  time = getlongtime();
  do {
    
    len = fio_write(&fio, buf, sizeof(buf));
    if (len != sizeof(buf)) {
	printf("Error writing %s (error %d, errno %d)\n", filename, r, errno);
	break;
    }

    //pos = fio.seek(fio.fh, 0, SEEK_CUR);
    //size = fio.seek(fio.fh, 0, SEEK_END);
    //fio.seek(fio.fh, pos, SEEK_SET);
    //printf("Pos: %d  Size: %d\n", pos, size);

    //printf(".");
    total += len;
    grand_total += len;

    if (getlongtime() >= time + 10000) {

	diff = getlongtime() - time;
	printf("[%lld kb] Wrote %lld kb in %lld msec - %lld kbps\n", grand_total/1024, total/1024, diff, diff ? total/diff : 0);
	time = getlongtime();
	total = 0;
    }

  } while (1);

  fio_close(&fio);

  if (len < 0)
    printf("THREAD %x ABORTED after %d bytes read, err = %d\n", pthread_self(), total, errno);
  else
    printf("THREAD %x EXITING after %d bytes read (eof), err = %d\n", pthread_self(), total, errno);

  return 0;
}

int show_attributes(char *fname)
{
    struct stat st;
    struct tm   tm;

    if (fio_init_smb() < 0)
	return -1;

    if (smbc_stat(fname, &st) < 0) {
	printf("Couldn't stat %s\n", fname);
	return -1;
    }

    printf("smbc_stat of %s:\n", fname);
    localtime_r(&st.st_ctime, &tm);
    printf("st_ctime = %s", asctime(&tm));
    localtime_r(&st.st_mtime, &tm);
    printf("st_mtime = %s", asctime(&tm));
    localtime_r(&st.st_atime, &tm);
    printf("st_atime = %s", asctime(&tm));
    return 0;
}

int dirlist(char *s)
{
    int fd;
    char name[256];
    struct smbc_dirent *de;

    fio_init_smb();

    if (!s || !s[0])
	sprintf(name, "smb://");
    else
	sprintf(name, "smb://%s", s);

    printf("Opening %s\n", name);
    fd = smbc_opendir(name);
    if (fd < 0) {
	printf("Couldn't open dir\n");
	return -1;
    }

    while (de = smbc_readdir(fd)) {
	printf("%s\n", de->name);
    }
    smbc_closedir(fd);
    return 0;
}

#pragma pack(1)
typedef struct _BITMAPFILEHEADER {
  short  bfType;
  long bfSize;
  short  bfReserved1;
  short  bfReserved2;
  long bfOffBits;
} __attribute__ ((packed)) BITMAPFILEHEADER, *PBITMAPFILEHEADER;
#pragma pack()

struct my_stat {
	unsigned short st_dev;
	unsigned short __pad1;
	unsigned long st_ino;
	unsigned short st_mode;
	unsigned short st_nlink;
	unsigned short st_uid;
	unsigned short st_gid;
	unsigned short st_rdev;
	unsigned short __pad2;
	unsigned long  st_size;
	unsigned long  st_blksize;
	unsigned long  st_blocks;
	unsigned long  st_atime;
	unsigned long  __unused1;
	unsigned long  st_mtime;
	unsigned long  __unused2;
	unsigned long  st_ctime;
	unsigned long  __unused3;
	unsigned long  __unused4;
	unsigned long  __unused5;
};

int main(int argc, char *argv[])
{
  int fh;
  pthread_t id;
  pthread_attr_t attr;
  struct stat st;

#if 0
  //printf("pad0 = %d\n", FIELD_OFFSET(struct stat, __pad0)); 
  printf("pad1 = %d\n", FIELD_OFFSET(struct stat, __pad1)); 
  printf("ino = %d\n", FIELD_OFFSET(struct stat, st_ino));
  printf("nlink = %d\n", FIELD_OFFSET(struct stat, st_nlink));
  printf("mode = %d\n", FIELD_OFFSET(struct stat, st_mode));
  printf("total = %d\n", sizeof(struct stat));
  printf("st_size = %d\n", sizeof(st.st_size));
  printf("st_blocks = %d\n", sizeof(st.st_blocks));
  return 0;
#endif
  printf("SMBTEST\n");
  printf("Copyright (c) 2005-2006 SkipJam Corp. ALL RIGHTS RESERVED.\n\n");
  printf("%d\n", sizeof(BITMAPFILEHEADER));

  if (argc < 3)
      return dirlist(NULL);
  else if (!strcasecmp(argv[1], "-a"))
      return show_attributes(argv[2]);
  else if (!strcasecmp(argv[1], "-d"))
      return dirlist(argv[2]);
  else if (!strcasecmp(argv[1], "-r"))
      return smbtest_thread(argv[2]);
  else if (!strcasecmp(argv[1], "-t"))
      return smbtest_test(argv[2]);
  else if (!strcasecmp(argv[1], "-w"))
      return smbtest_write_thread(argv[2]);
  else
      printf("Usage: smbtest [-d directory] [-f filename] [-a filename]\n");

#if 0
  if (argc < 2)
    return 0;

  pthread_mutex_init(&g_smbc_mutex, NULL);

  if (argc > 2) {
    
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);  //on it's own / auto clean when done
    pthread_create(&id, &attr, smbtest_thread, argv[2]);
  }

  //usleep(1000*1000);
  smbtest_thread(argv[1]);
#endif

  return 1;
}
