/*
 *  Embedded Ethernet driver for TANGO2 ES4 or above.
 */
#ifdef BOOTLOADER
#include "config.h"
#include "version.h"
#include "util.h"
#include "net.h"
#include "net_ipv4.h"
#include "em86xxapi.h"
#include "irqs.h"
#include "errno.h"
#include "uart.h"
#include "memcfg.h"
#else
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/compatmac.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/mii.h>
#include <linux/crc32.h>
#ifndef CONFIG_TANGO2
#include <asm/virtconvert.h>
#include <asm/arch/irqs.h>
#include <asm/arch/memcfg.h>
#else
#include <asm/tango2/memcfg.h>
#endif

// Use kernel timer implementation for polling the link status
//#define USE_KERNEL_TIMER

#endif //BOOTLOADER

#include "em86xx_eth.h"

/* Uncomment this for different messages */
//#define ETH_DEBUG 
#define ETH_ERRMSG 
#define ETH_MSG

#ifdef BOOTLOADER
#define DRIVER "em86xx_eth"

/* For static buffer allocation */
// #define STATIC_BUF_ALLOC
#define printk          uart_printf

#define PAGE_SHIFT	12
#define PAGE_SIZE	(1UL << PAGE_SHIFT)
#define PAGE_MASK	(~(PAGE_SIZE-1))
#define MAX_ORDER	10

#endif /* BOOTLOADER */

/* For software filtering*/
//#define USE_SW_FILTERING
#define USE_HW_FILTERING

/* For multi-packets frame support */
#define MULTI_PACKETS_FRAME_SUPPORT

#ifdef ETH_DEBUG
#define DBG_PRINT	printk
#else
static void inline DBG_PRINT(const char *x, ...) { ; }
#endif /* ETH_DEBUG */

#ifdef ETH_ERRMSG
#define ERR_PRINT	printk
#else
static void inline ERR_PRINT(const char *x, ...) { ; }
#endif /* ETH_ERRMSG */

#ifdef ETH_MSG
#define MSG_PRINT	printk
#else
static void inline MSG_PRINT(const char *x, ...) { ; }
#endif /* ETH_ERRMSG */

/* Hack: this is the mac address by default */
#define DEF_MAC_HI	0x00000011
#ifdef BOOTLOADER
#define DEF_MAC_LO	0xdeadbeef
#else
#define DEF_MAC_LO	0xaabbccdd
#endif

/* Default maximum link wait */
#ifdef USE_KERNEL_TIMER
#define DEF_LINK_LOOP		0x10000
#else
#define DEF_LINK_LOOP		0x100
#endif
#define DEF_MAX_LINK_LOOP	(DEF_LINK_LOOP * 16)

#define ETH_IRQ_FLAGS	(DmaIntNormal	  | DmaIntAbnormal    | DmaIntRxStopped		| \
			 DmaIntRxNoBuffer | DmaIntRxCompleted | DmaIntTxCompleted)

#define ETH_RXTX_FLAGS	(DmaTxStart       | DmaRxStart	      | DmaStoreAndForward)

#ifndef USE_HW_FILTERING 
#define ETH_MAC_FLAGS   (MacTxEnable      |  MacRxEnable       | MacPromiscuousModeOn	| \
                        MacPadStripEnable |  MacHashFilterOn   | MacPerfectFilterOff 	| \
                        MacRetryEnable    |  MacHeartBeatOff)
#else 
#define ETH_MAC_FLAGS	(MacTxEnable   	  | MacRxEnable        | MacPromiscuousModeOff  | \
			MacPadStripEnable | MacHashFilterOn    | MacPerfectFilterOn     | \
                        MacRetryEnable    | MacMulticastFilterOn | MacHeartBeatOff)
#define DEFAULT_BROADCAST_ADDRESS { 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF }
static void em86xx_eth_set_multicast_list(struct net_device *dev);
#endif /*USE_HW_FILTERING*/

#define MP_FRAME_FLAGS  (DescRxLast        | DescRxFirst)
#define TX_DESC1_FLAGS 	(DescTxLast        | DescTxFirst        | DescChain             | \
			 DescTxIntEnable)
#define RX_ERROR_FLAGS  (DescFilteringFail | DescError		| DescRxTruncated	| \
			 DescLengthError   | DescRxDamaged 	| DescRxLongFrame	| \
			 DescRxCollision   | DescRxMiiError	| DescRxCrc)
#define TX_ERROR_FLAGS  (DescTxLostCarrier | DescTxNoCarrier	| DescTxLateCollision 	| \
			 DescTxExcCollisions| DescTxHeartbeatFail| DescTxUnderflow)

#if defined(CONFIG_TANGO2) 
#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
extern int tango2_ethernet_getmac(unsigned char *mac);
#endif
#endif

#define PHY_10BASET_HD_ANAR  	 (CSMACD | BASET10 | FLOWCONTROL | RF)
#define PHY_10BASET_FD_ANAR 	 (CSMACD | BASET10 | BASET10FD | FLOWCONTROL | RF)
#define PHY_100BASET_HD_ANAR	 (CSMACD | BASET10 | BASET10FD | BASET100 | FLOWCONTROL | RF)
#define PHY_100BASET_FD_ANAR	 (CSMACD | BASET10 | BASET10FD | BASET100 | BASET100FD | FLOWCONTROL | RF)

/* Number of descriptors and buffer size */
#define MIN_NUM_RDESC	8
#define MIN_NUM_TDESC	8
#define NUM_RDESC 	32	
#define NUM_TDESC	32	
#define R_BUF_SIZE	0x700	/* needs to be < 2KB */
#define T_BUF_SIZE	0x700	/* needs to be < 2KB */

/* Private data for this ethernet device */
typedef struct {
	struct net_device_stats stats; /* stats */
	unsigned long last_rxidx;      /* last rx idx to descriptor */
	unsigned long next_txidx;      /* next tx idx to descriptor */
} EM86XX_ETH_PRIV;

/* Descriptors and buffers: notice the alignments */
#ifdef STATIC_BUF_ALLOC 
static struct em86xx_desc eth_rxdsc[NUM_RDESC] __attribute__ ((__aligned__(16)));
static struct em86xx_desc eth_txdsc[NUM_TDESC] __attribute__ ((__aligned__(16)));
static unsigned char eth_rxbuf[NUM_RDESC * R_BUF_SIZE] __attribute__ ((__aligned__(16)));
static unsigned char eth_txbuf[NUM_TDESC * T_BUF_SIZE] __attribute__ ((__aligned__(16)));
#else
static struct em86xx_desc *eth_rxdsc 	= NULL;
static struct em86xx_desc *eth_txdsc 	= NULL;
static unsigned char *eth_rxbuf 	= NULL;
static unsigned char *eth_txbuf		= NULL;
static unsigned long desc_page  	= 0L;
static unsigned long rxbuf_pages 	= 0L;
static unsigned long txbuf_pages 	= 0L;
#ifndef BOOTLOADER
static unsigned long rxbuf_order 	= 0L;
static unsigned long txbuf_order 	= 0L;
#endif
#endif

/* Number of descriptors */
static unsigned long num_rxdesc = NUM_RDESC;
static unsigned long num_txdesc = NUM_TDESC;

/* The real data pointer to be accessed in non-cache region */
static volatile struct em86xx_desc *rxdsc = NULL;
static volatile struct em86xx_desc *txdsc = NULL;
static volatile unsigned char *rxbuf	  = NULL;
static volatile unsigned char *txbuf	  = NULL;

/* Reference count for device */
static int dev_count = 0;

/* For storing MAC address */
static unsigned long def_mac_hi = DEF_MAC_HI;
static unsigned long def_mac_lo = DEF_MAC_LO;

/* Default PHY address: -1 for search */
static int phy_address = -1;
/* Default to 100BaseT/FD*/
static int phy_mode    = 0;

/* Link status flag */
static volatile int link_up 		 = 0;
static volatile int link_configured	 = 0;
static unsigned long max_link_loop	 = 0;
static unsigned long ld_count		 = 0;
static int reset_flag			 = 0;
static int phy				 = 0;

static int em86xx_eth_reset_desc(struct net_device *dev, int *reset);

#ifdef BOOTLOADER
static void em86xx_eth_intr_handler(int irq, void *dev_id);
#else
#ifdef USE_KERNEL_TIMER 
static struct timer_list eth_timer;
#else
static int stop_thread = 0;
#endif
static struct net_device em86xx_eth_dev;
static void em86xx_eth_intr_handler(int irq, void *dev_id, struct pt_regs *regs);
static struct sk_buff **rx_skb_list = NULL;
static int need_restart_tx_queue = 0; 

MODULE_AUTHOR("Craig Qu");
MODULE_DESCRIPTION("Sigma Designs ethernet driver");
MODULE_LICENSE("GPL");

MODULE_PARM(def_mac_hi, "i");
MODULE_PARM(def_mac_lo, "i");
MODULE_PARM(phy_address, "i");
MODULE_PARM(phy_mode,    "i");
MODULE_PARM_DESC(phy_address, "PHY address connecting to the MAC.");
MODULE_PARM_DESC(phy_mode, "Speed and duplex mode: 0=100BaseT/FD; 1=100BasT/HD; 2=10BaseT/FD; 3=10BaseT/HD");

#ifndef STATIC_BUF_ALLOC 
MODULE_PARM(num_rxdesc, "i");
MODULE_PARM(num_txdesc, "i");
#endif
#endif /* BOOTLOADER */

#define TX_ERRORS       12
#define TX_ERRORS_START 1

#define RX_ERRORS       31
#define RX_ERRORS_START 1

static const char *rx_error_msg[RX_ERRORS]= {
	NULL,
	"A CRC error is detected.",
	NULL,
	"A receive error was detected during frame reception.",
	NULL,NULL,
	"The frame has seen a collision after the collision window.",
	"The frame is longer than 1518 bytes.",
	NULL,NULL,NULL,
	"The frame is prematurely terminated before the collision window.",	
	"The actual length does not match with the Length Type field of the incoming frame.",
	NULL,
	"No more descriptors for receive frame.",
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	"The frame failed the address recognizing filtering.",
};

static const int rx_stat_offset[RX_ERRORS] = {
	-1,
	(int)&(((struct net_device_stats *)0)->rx_crc_errors),
	-1,-1,-1,-1,
	(int)&(((struct net_device_stats *)0)->collisions),
	-1,-1,-1,-1,-1,
	(int)&(((struct net_device_stats *)0)->rx_length_errors),
	-1,
	(int)&(((struct net_device_stats *)0)->rx_over_errors),
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
}; 

static const char *tx_error_msg[TX_ERRORS]= {
	NULL,
	"Late data arrival from the memory.",
	"Excessive deferral.",
	NULL,NULL,NULL,NULL,
	"Heartbeat collision check failure.",
	"Transmission aborted after 16 collisions.",
	"Transmission aborted due to collision.",
	"No carrier signal from the tranceiver.",
	"Loss of carrier during transmission.",	
};

static const int tx_stat_offset[TX_ERRORS] = {
	-1,
	(int)&(((struct net_device_stats *)0)->tx_aborted_errors),
	(int)&(((struct net_device_stats *)0)->tx_aborted_errors),
	-1,-1,-1,-1,
	(int)&(((struct net_device_stats *)0)->tx_heartbeat_errors),
	(int)&(((struct net_device_stats *)0)->collisions),
	(int)&(((struct net_device_stats *)0)->collisions),
	(int)&(((struct net_device_stats *)0)->tx_carrier_errors),
	(int)&(((struct net_device_stats *)0)->tx_carrier_errors),
}; 

/* Getting MAC address from local RAM */
static int get_mac_address(unsigned long *mac_lo, unsigned long *mac_hi)
{
	unsigned long cksum, lo, hi;
	const unsigned char marker = 'M';
	
	lo = em86xx_read_reg(REG_BASE_cpu_block + LR_ETH_MAC_LO);
	hi = em86xx_read_reg(REG_BASE_cpu_block + LR_ETH_MAC_HI);

	if (((unsigned char)((hi >> 16) & 0xff)) != marker)
		return(1); /* Not valid MAC */
	cksum = ((lo & 0xff) + ((lo >> 8) & 0xff) + ((lo >> 16) & 0xff) + ((lo >> 24) & 0xff) + 
		 (hi & 0xff) + ((hi >> 8) & 0xff)) & 0xff;
	if (((hi >> 24) & 0xff) != cksum)
		return(1); /* Not valid MAC */
	*mac_lo = lo;
	*mac_hi = hi & 0xffff;
	return(0); /* Got valid MAC */
}

/* Get the device mac address */
static int em86xx_get_macaddr(unsigned char *addr)
{
#if defined(CONFIG_TANGO2) && (defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV))
	unsigned long hi, lo;
	if (get_mac_address(&lo, &hi) == 0) {
		*addr++ = (hi & 0x0000ff00) >> 8;
		*addr++ = (hi & 0x000000ff);
		*addr++ = (lo & 0xff000000) >> 24;
		*addr++ = (lo & 0x00ff0000) >> 16;
		*addr++ = (lo & 0x0000ff00) >> 8;
		*addr   = (lo & 0x000000ff);
	} else if (tango2_ethernet_getmac(addr) == 0) {
		*addr++ = (def_mac_hi & 0x0000ff00) >> 8;
		*addr++ = (def_mac_hi & 0x000000ff);
		*addr++ = (def_mac_lo & 0xff000000) >> 24;
		*addr++ = (def_mac_lo & 0x00ff0000) >> 16;
		*addr++ = (def_mac_lo & 0x0000ff00) >> 8;
		*addr   = (def_mac_lo & 0x000000ff);
	}
#else
#ifdef BOOTLOADER
	if(	mac_read(addr) == -1) {
		*addr++ = (def_mac_hi & 0x0000ff00) >> 8;
		*addr++ = (def_mac_hi & 0x000000ff);
		*addr++ = (def_mac_lo & 0xff000000) >> 24;
		*addr++ = (def_mac_lo & 0x00ff0000) >> 16;
		*addr++ = (def_mac_lo & 0x0000ff00) >> 8;
		*addr   = (def_mac_lo & 0x000000ff);
	}
#else
#ifdef CONFIG_SD_BOOTLOADER_INTEGRATION
	{
		unsigned long lo, hi;

		if (get_mac_address(&lo, &hi) == 0) {
			def_mac_hi = hi;
			def_mac_lo = lo;
		}
	}
#endif
	*addr++ = (def_mac_hi & 0x0000ff00) >> 8;
	*addr++ = (def_mac_hi & 0x000000ff);
	*addr++ = (def_mac_lo & 0xff000000) >> 24;
	*addr++ = (def_mac_lo & 0x00ff0000) >> 16;
	*addr++ = (def_mac_lo & 0x0000ff00) >> 8;
	*addr   = (def_mac_lo & 0x000000ff);
#endif /* BOOTLOADER */	
#endif /* CONFIG_TANGO2 && (CONFIG_TANGO2_SIG_BLOCK || CONFIG_TANGO2_XENV) */
	return 0;
}

/* Setting up MAC address of ethernet PHY and device data */
static int em86xx_set_mac(struct net_device *dev)
{
	unsigned long hi_mac, low_mac;

	/* Set up device mac address */
	if (em86xx_get_macaddr(dev->dev_addr))
		return(-EIO);

	if (!is_valid_ether_addr(dev->dev_addr))
		MSG_PRINT("%s: bogus mac address detected.\n", dev->name);

	hi_mac =  (dev->dev_addr[5] << 8)  | dev->dev_addr[4];
	low_mac = (dev->dev_addr[3] << 24) | (dev->dev_addr[2] << 16) | 
		  (dev->dev_addr[1] << 8)  | dev->dev_addr[0];

	/* Set up MAC address */
	em86xx_write_reg(EM86XX_MACAHR_REG, hi_mac);
	em86xx_write_reg(EM86XX_MACALR_REG, low_mac);

	return 0;
}

#ifndef BOOTLOADER
/* Setting customized mac address */
static int em86xx_eth_set_macaddr(struct net_device *dev, void *addr)
{
	unsigned long hi_mac, low_mac;

	/* Check if given address is valid ethernet MAC address */
	if (!is_valid_ether_addr(addr))
		return(-EIO);

	/* Turn off IRQ and stop receive/transmit */
	em86xx_write_reg(EM86XX_CR_REG, 0);
	em86xx_write_reg(EM86XX_IER_REG, 0);

	netif_carrier_off(dev); /* Shutdown interface */

	/* Save the customize mac address */
	memcpy(dev->dev_addr, addr, 6);
	hi_mac  = (dev->dev_addr[5] << 8) | dev->dev_addr[4];
	low_mac = (dev->dev_addr[3] << 24)| (dev->dev_addr[2] << 16) | 
		  (dev->dev_addr[1] << 8) | dev->dev_addr[0];

	def_mac_hi = hi_mac;
	def_mac_lo = low_mac;

	/* Set up MAC address */
	em86xx_write_reg(EM86XX_MACAHR_REG, hi_mac );
	em86xx_write_reg(EM86XX_MACALR_REG, low_mac );

	/* Restart everything */
	link_up = link_configured = 0;

	return 0;
}
#endif /* BOOTLOADER */

/* RTL8201BL is a single-port PHY with an MII/SNI. The MAC layer device can 
   control a maximum of 31 devices, configured with different PHY addresses
   (00001b to 11111b). We need to find out which PHY address is using in 
   order to obtain the mode status after the auto-neotiation is completed.

   There is a module parameter phy_address to be used to override this probe,
   to set PHY address if one already knew it.
*/
static int em86xx_phy_probe(void)
{

        int phy_addr 	= 0;
	u16 phy_id 	= 0;
        u16 phy_status 	= 0;

	if (phy_address < 0) {
		/* search for total of 31 possible mii phy addresses */
		for (phy_addr = 1; phy_addr < 32; phy_addr++) {
			phy_status = em86xx_mii_read(phy_addr,GEN_sts);

			if (phy_status != 0xffff && phy_status != 0x0000)
				break;
		}
	} else {
		phy_addr = phy_address;
		phy_status = em86xx_mii_read(phy_addr,GEN_sts);
	}
	phy_id =  em86xx_mii_read(phy_addr,GEN_id_lo);
	MSG_PRINT("Found PHY %04x at Addr = %d Status=0x%x\n", phy_id, phy_addr, phy_status);
 
	return phy_addr;
}

/* Initialize the ethernet link status, mac, and flow control */
static int em86xx_link_config(struct net_device *dev, int timeout)
{
	unsigned long word = 0;
 	int i = timeout, fdx = 0, speed = 0;
	static int msg_cnt = 0;

	/* If link is up and configured, don't do anything */
	if ((link_up != 0) && (link_configured != 0))
		return 0;
	else
		DBG_PRINT("%s: re-configure link (%d).\n", dev->name, i);

	/* Turn off IRQ and stop receive/transmit */
	em86xx_write_reg(EM86XX_CR_REG, 0);
	em86xx_write_reg(EM86XX_IER_REG, 0);

        if( (em86xx_mii_read(phy, GEN_sts) & LINK) == 0 ) {
		if ((++msg_cnt % 30) == 1)
			MSG_PRINT( "em86xx_eth::open(%s) - no link\n", dev->name );
		return -ENODEV;
	} else
		msg_cnt = 0;

	/* Reset phy */
        em86xx_mii_write(phy, GEN_ctl, RESET);
        em86xx_mii_write(phy, GEN_ctl, 0);

	/* Set phy mode */
	switch(phy_mode){
		case 1: em86xx_mii_write(phy, AN_adv, PHY_100BASET_HD_ANAR);
			break;
		case 2: em86xx_mii_write(phy, AN_adv, PHY_10BASET_FD_ANAR);
			break;
		case 3: em86xx_mii_write(phy, AN_adv, PHY_10BASET_HD_ANAR);
			break;
		default:em86xx_mii_write(phy, AN_adv, PHY_100BASET_FD_ANAR);
			break;
	}
	
	/* Restart auto-negotiation */
        em86xx_mii_write(phy, GEN_ctl, AUTONEG_ENB | AUTONEG_REST);

        /* Wait for autonegotiation complete */
        while( i-- > 0 ) { 
#ifdef USE_KERNEL_TIMER
		udelay(1); 
#else
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ/10);
#endif
                if( (em86xx_mii_read(phy, GEN_sts) & AUTOCMPLT) != 0 )
			break;
        }

	if (i <= 0) {
		MSG_PRINT("%s: auto-negotiation failed.\n", dev->name);
		return -ENODEV;
	} else
		link_up = 1;

        fdx = (em86xx_mii_read(phy, GEN_ctl) & DUPLEX) != 0;
        MSG_PRINT(  "em86xx_eth::open(%s) - %s-duplex mode\n",
                     dev->name, fdx ? "Full" : "Half" );

	/* RTL 8201BL Register 25 Test Register for checking Link status*/
        /*speed = (em86xx_mii_read(phy, GEN_ctl) & SPEED) != 0;*/
        speed = em86xx_mii_read(phy, TST) & (LINK10 | LINK100);
        switch(speed) {
                case 0:  
                        return -ENODEV;
                case 1: 
                        speed = 1;
                        break;
                case 2: 
                        speed = 0;
                        break;
                default:
			speed = 1; 
                        break;
        }

        MSG_PRINT(  "em86xx_eth::open(%s) - 10%s Mbit/s\n",
                           dev->name, speed ? "0" : "" );

	/* set the descriptor base address */
	em86xx_write_reg(EM86XX_RLBAR_REG, PHYSADDR((u32)rxdsc) );
	em86xx_write_reg(EM86XX_TLBAR_REG, PHYSADDR((u32)txdsc));

	/* set bus mode: burst length = 32 | BLE | ~DBO */
	word = (32 << 8) /*| (1 << 7) | (1 << 20)*/ ;
	em86xx_write_reg(EM86XX_BMR_REG, word); 

	/* enable MAC flow ctrl */
	word = (1 << 1);
	em86xx_write_reg(EM86XX_FCR_REG, word); 

	/* configure MAC ctrller */
	word = ETH_MAC_FLAGS;
	word |= MacLoopbackOff; 
	if (fdx)
  		word |= MacFullDuplex;    
	em86xx_write_reg(EM86XX_MACCR_REG, word);

	/* Turn on the IRQ and start receive/transmit */
	em86xx_write_reg(EM86XX_IER_REG, ETH_IRQ_FLAGS);
	em86xx_write_reg(EM86XX_CR_REG, ETH_RXTX_FLAGS); 

	link_configured = 1;

	return 0;
}

/* Ethernet hardware initialization */
static int em86xx_eth_hw_init(struct net_device *dev)
{
	/* reset dma engine*/
	em86xx_write_reg(EM86XX_BMR_REG, DmaResetOn);
	em86xx_write_reg(EM86XX_BMR_REG, DmaResetOff);

	if (em86xx_set_mac(dev))
		return(-EIO);

	/* stop MAC engine */
	em86xx_write_reg(EM86XX_CR_REG, 0x0); 
#ifndef USE_HW_FILTERING
	/* set up multicast hash table to MCHTHR MCHTLR */
	/* set multicast hash table to accept all */
	em86xx_write_reg(EM86XX_MCHTHR_REG, 0xffffffff); 
	em86xx_write_reg(EM86XX_MCHTLR_REG, 0xffffffff); 
#else
	/* clear hash table */
        em86xx_write_reg( EM86XX_MCHTHR_REG, 0 );
        em86xx_write_reg( EM86XX_MCHTLR_REG, 0 );

	em86xx_eth_set_multicast_list(dev);
#endif

	/* resetting descriptors */
#ifdef BOOTLOADER
        if (em86xx_eth_reset_desc(dev, &reset_flag))
#else
        if (em86xx_eth_reset_desc(&em86xx_eth_dev, &reset_flag))
#endif
		return(-EIO);

	/* configure PHY and speed */
	return em86xx_link_config(dev, max_link_loop);
}

#ifndef BOOTLOADER
/* Monitor the status of link, re-do link initialization if necessary. */
#ifdef USE_KERNEL_TIMER
static void em86xx_eth_link_monitor(unsigned long dummy)
#else
static int em86xx_eth_link_monitor(void *dummy)
#endif
{
	unsigned long link_status;
	struct net_device *dev = (struct net_device *)dummy;
	int res;

#ifndef USE_KERNEL_TIMER
	strcpy(current->comm, "em86xx_eth");
	daemonize();
	reparent_to_init();

	while (stop_thread == 0) {
#endif

	/* Setup the link/configuration is not done so. */
	if ((link_up == 0) || (link_configured == 0)) {
		res = em86xx_eth_hw_init(dev);
		if (res < 0)
			goto out;
	}

	/* Get link status and adjust the delay loop accordingly */
	link_status = (em86xx_mii_read(phy, GEN_sts) & LINK) ;
	if (link_status && (link_status != 0xffff)) {
  		if ((link_configured != 0) && (link_up != 0)) {
    			if (!netif_carrier_ok(dev)) {
	      			netif_carrier_on(dev); /* Up and running, turn netif on */
	      	        	max_link_loop = DEF_LINK_LOOP;
	      	        	ld_count >>= 2;

        	      		if (ld_count == 0)
      					ld_count = 1;
	    		}
	        } else {
	      		ld_count <<= 1;
	      		max_link_loop = ld_count * DEF_LINK_LOOP;
	
	      		if (max_link_loop > DEF_MAX_LINK_LOOP)
	      			max_link_loop = DEF_MAX_LINK_LOOP;
	
	              	netif_carrier_off(dev);
	              	link_configured = 0;
	     	}
	} else {
		MSG_PRINT("%s: detected link down.\n", dev->name);
	        ld_count <<= 1;
	        max_link_loop = ld_count * DEF_LINK_LOOP;

	  	if (max_link_loop > DEF_MAX_LINK_LOOP)
    			max_link_loop = DEF_MAX_LINK_LOOP;

	  	netif_carrier_off(dev);
	  	link_up = link_configured = 0;
	}

out:
#ifdef USE_KERNEL_TIMER
	/* Schedule for the next time */
	mod_timer(&eth_timer, jiffies + HZ); 
#else
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ);
	}
	return(0);
#endif
}
#endif

/* Setting rx/tx descriptors */
static void em86xx_eth_setup_desc(void)
{
	register int i;
	struct em86xx_desc *desc_ptr = NULL;
	unsigned long base_addr = 0;

	/* Setup rx descriptor */
	desc_ptr = (struct em86xx_desc *)(rxdsc);
	for (i = 0; i < (num_rxdesc - 1); i++, desc_ptr++) {
		desc_ptr->desc3 = PHYSADDR((unsigned long)(desc_ptr + 1));
		desc_ptr->desc1 = (DescChain | R_BUF_SIZE);
  		desc_ptr->desc0 = DescOwnByDma; 
	}
	desc_ptr->desc3 = PHYSADDR((unsigned long)(rxdsc));
	desc_ptr->desc1 = (DescChain | DescEndOfRing | R_BUF_SIZE);
	desc_ptr->desc0 = DescOwnByDma; 

	/* Setup tx descriptor */
	desc_ptr = (struct em86xx_desc *)(txdsc); 
	for (i =0;  i < (num_txdesc - 1); i++, desc_ptr++) {
		desc_ptr->desc3 = PHYSADDR((unsigned long)(desc_ptr + 1));
		desc_ptr->desc1 = (TX_DESC1_FLAGS | T_BUF_SIZE);
  		desc_ptr->desc0 = DescOwnByCPU;
	}
	desc_ptr->desc3 = PHYSADDR((unsigned long)(txdsc));
	desc_ptr->desc1 = (TX_DESC1_FLAGS | DescEndOfRing | T_BUF_SIZE);
	desc_ptr->desc0 = DescOwnByCPU;

	/* Point rx descriptor to buffer */
	desc_ptr = (struct em86xx_desc *)(rxdsc);
#ifdef BOOTLOADER 
      	base_addr = PHYSADDR((unsigned long)(rxbuf)); 
	for (i = 0; i < num_rxdesc; i++, desc_ptr++, base_addr += R_BUF_SIZE) 
		desc_ptr->desc2 = base_addr;
#else
        for (i = 0; i < num_rxdesc; i++, desc_ptr++){
                rx_skb_list[i] = dev_alloc_skb(R_BUF_SIZE);
                skb_reserve(rx_skb_list[i], 2);
                desc_ptr->desc2 = PHYSADDR((u32)rx_skb_list[i]->data);
#ifdef CONFIG_NONCOHERENT_IO
		dma_cache_inv((u32)rx_skb_list[i]->data, R_BUF_SIZE);
#endif
        }
#endif

	/* Point tx descriptor to buffer */
	desc_ptr = (struct em86xx_desc *)(txdsc);
        base_addr = PHYSADDR((unsigned long)(txbuf));
	for (i = 0; i < num_txdesc; i++, desc_ptr++, base_addr += T_BUF_SIZE) 
		desc_ptr->desc2 = base_addr;

}

/* Starting up the ethernet device */
static int em86xx_eth_open(struct net_device *dev)
{
	EM86XX_ETH_PRIV *private = (EM86XX_ETH_PRIV *)dev->priv;

	if ((dev == NULL) || (private == NULL))
		return(-EIO);
	DBG_PRINT("%s: starting interface.\n", dev->name);

	/*Get phy address*/
	em86xx_write_reg(EM86XX_CR_REG, 0);
	em86xx_write_reg(EM86XX_IER_REG, 0);
	phy = em86xx_phy_probe();
	em86xx_eth_hw_init(dev);

#ifdef BOOTLOADER
        dev->state = NETDEV_UP;
#else
#ifdef USE_KERNEL_TIMER 
	/* Schedule timer for monitoring link status */
	init_timer(&eth_timer);
	eth_timer.function = em86xx_eth_link_monitor;
	eth_timer.data = (unsigned long)dev;
	mod_timer(&eth_timer, jiffies + HZ/10);
#else
	stop_thread = 0;
	if (kernel_thread(em86xx_eth_link_monitor, dev, CLONE_FS|CLONE_FILES) < 0)
		return(-ENODEV);
#endif
	netif_start_queue(dev);
#endif /*BOOTLOADER*/

	return 0;
}

/* Stopping the ethernet device */
static int em86xx_eth_close(struct net_device *dev)
{
	if (dev == NULL)
  		return(-EIO);

	/* Turn off IRQ and stop receive/transmit */
	em86xx_write_reg(EM86XX_CR_REG, 0);
	em86xx_write_reg(EM86XX_IER_REG, 0);

#ifdef BOOTLOADER
	dev->state = NETDEV_DOWN;
#else

#ifdef USE_KERNEL_TIMER
	/* Kill timer */
	del_timer_sync(&eth_timer);
#else
	stop_thread = 1;
	schedule_timeout(5*HZ); /* Wait for kernel thread to terminate */
#endif

	/* Stop the transmit queue */
	netif_stop_queue(dev);
	link_up = link_configured = 0;

#endif /*BOOTLOADER*/
	return 0;
}

#ifndef BOOTLOADER
#ifdef USE_HW_FILTERING
/* Setup multicast list */
static void em86xx_eth_set_multicast_list(struct net_device *dev)
{
	/* Multicast hash filter */
        u32 mc_filter[2] = { 0, 0 };
        int i, rx_mode;
 	u8  broadcast_addr[6]  = DEFAULT_BROADCAST_ADDRESS;

	rx_mode =  em86xx_read_reg(EM86XX_MACCR_REG);  
        DBG_PRINT(KERN_DEBUG"%s:   set_rx_mode(0x%x) done -- rx_mode=0x%x.\n",
                           dev->name, dev->flags, rx_mode);

        if (dev->flags & IFF_PROMISC) {
                MSG_PRINT("%s: Promiscuous mode enabled.\n", dev->name);
                rx_mode |= MacPromiscuousModeOn;
                mc_filter[1] = mc_filter[0] = 0xffffffff;
        } else {
                struct dev_mc_list *mclist;
		int n;
                rx_mode &=  ~MacPromiscuousModeOn;
                for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
                         i++, mclist = mclist->next){
			n = ether_crc(ETH_ALEN, mclist->dmi_addr) >> 26;
			mc_filter[n >> 5] |= 1 << (n & 31);
		}
		n = ether_crc(ETH_ALEN, broadcast_addr) >> 26;
		mc_filter[n >> 5] |= 1 << (n & 31);
        }
	em86xx_write_reg(EM86XX_MACCR_REG, rx_mode );
        em86xx_write_reg(EM86XX_MCHTLR_REG, mc_filter[0]);
        em86xx_write_reg(EM86XX_MCHTHR_REG, mc_filter[1]);

        return;
}
#endif
#endif

/* Transmit a packet */
#ifdef BOOTLOADER
static int em86xx_eth_tx(struct sk_buff *skb, struct net_device *dev, int async)
#else
static int em86xx_eth_tx(struct sk_buff *skb, struct net_device *dev)
#endif
{
	EM86XX_ETH_PRIV *private = (EM86XX_ETH_PRIV *)dev->priv;
	volatile struct em86xx_desc *desc_ptr;
	struct net_device_stats *stats;
	unsigned long length, err;

	if ((link_up == 0) || (link_configured == 0)) {
		DBG_PRINT("em86xx_eth_tx: link is not up yet\n");
#ifndef BOOTLOADER
		netif_carrier_off(dev);
#endif
		return(-EIO); 
	} 

	if ((length = ((ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN)) > T_BUF_SIZE) {
      		DBG_PRINT("em86xx_eth_tx: too big\n");
		return(-ENOMEM); 
	}

	stats = &(private->stats);

	/* Send packet to device */
	desc_ptr = (volatile struct em86xx_desc *)(&txdsc[private->next_txidx]);
	if ((desc_ptr->desc0 & DescOwnByDma) != 0) {
		stats->tx_dropped++;
		
		/* cannot queue anymore */
		netif_stop_queue(dev);
		need_restart_tx_queue = 1;

        	ERR_PRINT("%s desc_ptr=0x%x: tx error (descriptor not owned by CPU).\n",
				 dev->name, (u32)desc_ptr);
		return(-EIO);
	} 

	/* Check if previous transmission has error */
	if ((err = (desc_ptr->desc0 & TX_ERROR_FLAGS)) != 0) {
		int i;
		ERR_PRINT("%s: got a tx error:0x%08lx\n", dev->name, desc_ptr->desc0);
		for (i = TX_ERRORS_START; (i < TX_ERRORS) && (err != 0); i++) {
			int *ptr;
			if (err & (1 << i)) {
				err &= ~(1<<i);
				if (tx_error_msg[i] != NULL)
					ERR_PRINT("%s\n",tx_error_msg[i]);
				if (tx_stat_offset[i] >= 0) {
					ptr = (int *)(((char *)stats) + tx_stat_offset[i]);
					*ptr++;
				}
			}
		}

		/* Update stats */
 		stats->collisions += (desc_ptr->desc0 >> 3) & 15;
		stats->tx_errors++;
		stats->tx_packets--;
		stats->tx_bytes -= (desc_ptr->desc1 & DescSize1Mask);
	}

	/* Copy packet data to tx buffer */
	DBG_PRINT("tx copy data from skb->data=0x%x to desc2=0x%x length=0x%x\n",
		(u32)skb->data, (u32)desc_ptr->desc2, (u32)length );

	memcpy((void *)NON_CACHED(desc_ptr->desc2), skb->data, length);

	/* Setup tx descriptor */
	desc_ptr->desc1 = ( TX_DESC1_FLAGS | length );
	if (private->next_txidx == (num_txdesc - 1))
		desc_ptr->desc1 |=  DescEndOfRing ;
	desc_ptr->desc0 |= DescOwnByDma;

	/* Start transmission (should be in suspend mode already) */
	em86xx_write_reg(EM86XX_TPDR_REG, 0x1);

#ifdef BOOTLOADER
        private->next_txidx = imodulus((private->next_txidx + 1), num_txdesc);
#else
	private->next_txidx = ((private->next_txidx + 1) % num_txdesc);
	dev->trans_start = jiffies;
#endif
	/* Update stats */
	stats->tx_packets++;
	stats->tx_bytes += length;

	/* Free up socket buffer */
#ifdef BOOTLOADER
	async = 0;
	skb_free(skb);
#else
	dev_kfree_skb(skb);
#endif
	return 0;
}


static inline int is_sw_filtered_packet(struct net_device *dev, unsigned char *data)
{
#ifdef USE_SW_FILTERING
        if (((data[0] & 0x80) == 0) && (dev->dev_addr[5] != data[5]))
                return 1;
        else
                return 0;
#else
        return 0;
#endif
}


/* Receiving packet(s) */
static int em86xx_eth_rx(struct net_device *dev)
{
	EM86XX_ETH_PRIV *private;
#ifdef BOOTLOADER
	struct sk_buff *skb;
#endif
	struct net_device_stats *stats;
	unsigned char *data;
	volatile struct em86xx_desc *desc_ptr;
	unsigned long length;
	unsigned long dsc;
	unsigned long count=0, cnt= 0, err;
#ifdef MULTI_PACKETS_FRAME_SUPPORT
	long mpf_start = -1;
#endif

  	private = (EM86XX_ETH_PRIV *)dev->priv;
  	stats = &(private->stats);

	/* Go thru descriptors list until we ran out or encounterred the one
	   being processed by DMA */
#ifdef BOOTLOADER 
	for (cnt = 0, count = private->last_rxidx; cnt < num_rxdesc; cnt++, count = imodulus(count + 1, num_rxdesc)) {
#else
	for (cnt = 0, count = private->last_rxidx; cnt < num_rxdesc; cnt++, count = ((count + 1) % num_rxdesc)) {
#endif

		desc_ptr = (volatile struct em86xx_desc *)(&rxdsc[count]);
	 	dsc = desc_ptr->desc0;

	  	if ((dsc & DescOwnByDma) != 0) {
    			/* DMA is processing this one, break out the loop */
			break;
    		}

		/*filter out those are not for me */
      		data = (unsigned char *)NON_CACHED(desc_ptr->desc2);
		if (is_sw_filtered_packet(dev, data)) {
			ERR_PRINT("%s: mac mismatched, dropped.\n", dev->name);
		} else	if ((err = (dsc & RX_ERROR_FLAGS)) != 0)  {
			int i;
      			ERR_PRINT("%s: rx error 0x%08lx:%ld desc2=0x%08x\n", 
				dev->name, dsc, count, (u32)desc_ptr->desc2);

			for (i = RX_ERRORS_START; (i < RX_ERRORS) && (err != 0); i++) {
				int *ptr;
				if (err & (1 << i)) {
					err &= ~(1<<i);
					if (rx_error_msg[i] != NULL)
						ERR_PRINT("%s\n", rx_error_msg[i]);
					if (rx_stat_offset[i] >= 0) {
						ptr = (int *)(((char *)stats) + rx_stat_offset[i]);
						*ptr++;
					}
				}
			}

			/* We dropped any error packet */
		        stats->rx_errors++;
		 } else if ((dsc & MP_FRAME_FLAGS) != MP_FRAME_FLAGS) {
#ifdef MULTI_PACKETS_FRAME_SUPPORT
			 /* We encounterred multi-packets frame */
			 if (((dsc & DescRxFirst) != 0) && ((dsc & DescRxLast) == 0)) {
				 /* First packet of multi-packet frame */
				 mpf_start = count;
				 continue; /* Loop until we can get the whole frame */
			 } else if ((dsc & MP_FRAME_FLAGS) == 0) {
				 /* Interim packets within multi-packet frame */
				 continue; /* Loop until we can get the whole frame */
			 } else {
				 /* Last packet of multi-packet frame */
				 unsigned long mpf_size;
				 long i, mpf_end;
				 char *ptr, *mpf_data;
				 volatile struct em86xx_desc *mpf_desc_ptr;
				 struct sk_buff *mpf_skb;

				 /* First we get the size of the frame, and allocate a new skb */
				 mpf_size = (dsc & DescFrameLengthMask) >> DescFrameLengthShift;
				 mpf_end = (count + 1) % num_rxdesc;
#ifdef BOOTLOADER
				 mpf_skb = skb_alloc(mpf_size);
#else
				 mpf_skb = dev_alloc_skb(mpf_size);
#endif
				 if (mpf_skb == NULL) {
					 /* Error: no memory available, dropping all packets in this frame */
					 stats->rx_dropped += (((mpf_start + num_rxdesc) - mpf_end) % num_rxdesc);
					 for (i = mpf_start; i != mpf_end; i = (i + 1) % num_rxdesc) {
						 mpf_desc_ptr = (volatile struct em86xx_desc *)(&rxdsc[i]);
						 mpf_desc_ptr->desc0 = DescOwnByDma;
					 }
				 } else {
#ifndef BOOTLOADER
					 skb_reserve(mpf_skb, 2);
#endif
					 stats->rx_bytes += mpf_size;
					 stats->rx_packets += (((mpf_start + num_rxdesc) - mpf_end) % num_rxdesc);

					 /* Collecting all packets into skb  */
					 for (ptr = mpf_skb->data, i = mpf_start; i != mpf_end; i = (i + 1) % num_rxdesc) {
						 mpf_desc_ptr = (volatile struct em86xx_desc *)(&rxdsc[i]);
						 mpf_data = (unsigned char *)NON_CACHED(mpf_desc_ptr->desc2);
						 memcpy(ptr, mpf_data, (mpf_size > R_BUF_SIZE) ? R_BUF_SIZE : mpf_size);
						 ptr += R_BUF_SIZE;
						 mpf_size -= R_BUF_SIZE;
						 mpf_desc_ptr->desc0 = DescOwnByDma;
					 }
					 mpf_skb->dev = dev;
					 mpf_skb->len = mpf_size - 4;

					 /* Send the frame back up */
#ifdef BOOTLOADER
					 skb_put(mpf_skb);
#else
					 mpf_skb->protocol = eth_type_trans(mpf_skb, dev);
					 mpf_skb->ip_summed = CHECKSUM_NONE;
					 dev->last_rx = jiffies;
					 netif_rx(mpf_skb);
#endif
				 }
				 mpf_start = -1;
				 continue; /* No need go to the end of the loop */
			}
#else
			/* We encounterred multi-packets frame */
	        	ERR_PRINT("%s: multi-packets frame detected 0x%08lx\n", dev->name, dsc);
		        /* Don't handle multi-packets frame */
		        stats->rx_dropped++;
#endif /* MULTI_PACKETS_FRAME_SUPPORT */
		 } else if ((length = ((dsc & DescFrameLengthMask) >> DescFrameLengthShift)) > R_BUF_SIZE) {
			/* Should not happen for signle-packet frame */
      			ERR_PRINT("%s: rx dropped (size too large: %ld)\n", dev->name, length);
			stats->rx_length_errors++;	
			stats->rx_dropped++;
#ifdef BOOTLOADER
                 } else if ((skb = skb_alloc(length)) == NULL) {
			ERR_PRINT("%s: rx dropped (memory unavailable)\n", dev->name);
	        	stats->rx_dropped++;
#endif
		 } else {
			 /* We got single-packet frame */
#ifdef BOOTLOADER
		  	 skb->dev = dev;
                         skb->len = length - 4;
                         memcpy(skb->data, data, length-4);
                         skb_put(skb);

                         /* Update stats */
                         stats->rx_packets++;
                         stats->rx_bytes += length;
#else
                         rx_skb_list[count]->dev = dev;
                         rx_skb_list[count]->len = length-4;
                         rx_skb_list[count]->protocol = eth_type_trans(rx_skb_list[count], dev);
                         rx_skb_list[count]->ip_summed = CHECKSUM_NONE;

                         /* Send the packet to kernel */
                         netif_rx(rx_skb_list[count]);

	        	 /* Update stats */
		       	 stats->rx_packets++;
		         stats->rx_bytes += length;
		       	 dev->last_rx = jiffies;
#endif /*BOOTLOADER*/
  	 	}
#ifndef BOOTLOADER
                rx_skb_list[count] = dev_alloc_skb(R_BUF_SIZE);
                skb_reserve(rx_skb_list[count], 2);
                desc_ptr->desc2 = PHYSADDR((u32)rx_skb_list[count]->data);
#ifdef CONFIG_NONCOHERENT_IO
		dma_cache_inv((u32)rx_skb_list[count]->data, R_BUF_SIZE);
#endif
#endif
  		desc_ptr->desc0 = DescOwnByDma;
  	}

#ifdef MULTI_PACKETS_FRAME_SUPPORT
	/* This is the last desc we read */
	if (mpf_start >= 0)
		/* We haven't reached the end of multi-packets frame, mark it here
		 * so we can come back and pick it up again */
		private->last_rxidx = mpf_start; 
	else
		private->last_rxidx = count; 
#else
	/* This is the last desc we read */
	private->last_rxidx = count; 
#endif
	/* make sure rx is not suspended */
	em86xx_write_reg(EM86XX_RPDR_REG, 0x1);
  	return cnt;
}

/* Resetting rx descriptor: error recovery for "rx buf unavailable" */
static void em86xx_eth_reset_rx_buf(struct net_device *dev)
{
	register int i;
	volatile struct em86xx_desc *desc_ptr = NULL;

	em86xx_clear_reg(EM86XX_CR_REG, DmaRxStart); 

	/* Reset rx desc */
	desc_ptr = (struct em86xx_desc *)(rxdsc); 
	for (i = 0; i < num_rxdesc; i++, desc_ptr++) 
		desc_ptr->desc0 = DescOwnByDma; 
  	
  	em86xx_write_reg(EM86XX_RPDR_REG,  0x1);
	em86xx_set_reg(EM86XX_CR_REG, DmaRxStart); 
}

#ifndef BOOTLOADER
/* Get the stats information */
static struct net_device_stats *em86xx_eth_stats(struct net_device *dev)
{
	if (dev != NULL) 
    		return(&(((EM86XX_ETH_PRIV *)dev->priv)->stats));
  	else
    		return(NULL);
}
#endif

/* Ethernet interrupt handler */
#ifdef BOOTLOADER
static void em86xx_eth_intr_handler(int irq, void *dev_id)
#else
static void em86xx_eth_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
	int num_rx_desc_freed = 0;
	unsigned long status;
	struct net_device *dev = (struct net_device *)dev_id;

#ifdef BOOTLOADER
        if( memcmp(dev->name,DRIVER, sizeof(DRIVER)) != 0) 
                return;
#else
        if (dev_id != (void *)&em86xx_eth_dev)
                return;
#endif
        /* Check status */
        status = em86xx_read_reg(EM86XX_SR_REG);

        // clear all interrupt requests
        em86xx_write_reg(EM86XX_SR_REG, status) ;

        if (status & DmaIntNormal) {
		 
                if (status & DmaIntRxCompleted)  
	                num_rx_desc_freed = em86xx_eth_rx(dev);
                else if(status & DmaIntTxCompleted){
			if(need_restart_tx_queue){
				netif_start_queue(dev);
				need_restart_tx_queue = 0;
			}
		}
		else
                        ERR_PRINT("%s: unhandled NIS 0x%08lx\n", dev->name, status);
        } 
 
	if (status & DmaIntAbnormal) { 
                if (status & DmaIntRxNoBuffer) {
			if (num_rx_desc_freed == 0)
	                	num_rx_desc_freed = em86xx_eth_rx(dev);

			if(num_rx_desc_freed == 0){
                        	ERR_PRINT("%s: receive buffer unavailable 0x%08lx\n", dev->name, status);
                        	em86xx_eth_reset_rx_buf(dev);
			}
                } else if (status & DmaIntRxStopped) {
                        ERR_PRINT("%s: receive process stopped\n", dev->name);
                        em86xx_set_reg(EM86XX_CR_REG, DmaRxStart); 
                } else if (status & DmaIntTxStopped ) {
                        ERR_PRINT("%s: transmit process stopped\n", dev->name);
                        em86xx_write_reg(EM86XX_CR_REG, DmaTxStart); 
                } else if (status & DmaIntTxUnderflow) {
                        ERR_PRINT("%s: transmit underflow\n", dev->name);
                } else if (status & DmaIntEarlyTx ) {
                        ERR_PRINT("%s: Early transmit interrupt\n", dev->name);
                } else if (status & DmaIntBusError ) {
                        ERR_PRINT("%s: Fatal bus error\n", dev->name);
                } else {
                        ERR_PRINT("%s: unhandled AIS 0x%08lx\n", dev->name, status);
                }

        }
 
	if ((status & (DmaIntAbnormal | DmaIntNormal)) == 0) {
                DBG_PRINT("%s: Unknown SR 0x%08lx\n", dev->name, status);
        }

        return;
}

#ifndef BOOTLOADER
/* Handling ioctl call */
static int em86xx_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
 	int ret = 0;
	struct ifr_data_struct
 	{
		u32 unit;
    		u32 addr;
    		u32 data;
  	} *req = (struct ifr_data_struct *)ifr->ifr_data;

 	switch( cmd )
 	{
		case SIOCDEVPRIVATE:  
			DBG_PRINT( "em86xx::read registers(%s, cmd=%08x) unit=0x%x  reg=0x%x\n", 
				    dev->name, cmd, req->unit, req->addr );

			if( req->unit == 0 ){        /* Read register */
      				req->data = em86xx_read_reg( req->addr );
    			}
			else if( req->unit == 1 ){   /* Read MAC register */
      				req->data = em86xx_read_mac_reg( req->addr );
    			}
    			else if( req->unit == 2 ) {   /* Read DMA register */
      				req->data = em86xx_read_dma_reg( req->addr );
    			}
    			else if( req->unit == 3 ) {   /* Read MII register */
      				req->data = em86xx_mii_read( phy, req->addr );
    			}
    			else
				 ret = -EOPNOTSUPP;
    			break;

  		case SIOCDEVPRIVATE+1:                /* Write registers */
			DBG_PRINT( "em86xx::write registers(%s, cmd=%08x) unit=0x%x reg=0x%x data=0x%x\n", 
				    dev->name, cmd ,req->unit, req->addr, req->data);

    			if( req->unit == 0 ) {        /* Write register */
      				em86xx_write_reg( req->addr, req->data );
    			}
    			else if( req->unit == 1 ) {        /* Write MAC register */
      				em86xx_write_mac_reg( req->addr, req->data );
    			}
    			else if( req->unit == 2 ) {   /* Write DMA register */
      				em86xx_write_dma_reg( req->addr, req->data );
    			}
    			else if( req->unit == 3 ) {   /* Write MII register */
      				em86xx_mii_write( phy, req->addr, req->data );
    			}
    			else 
				ret = -EOPNOTSUPP;
    			break;

 		default:
    			ret = -EOPNOTSUPP;
  	}

	return ret;
}
#endif

/* Kernel level ethernet initialization */
static int em86xx_eth_init(struct net_device *dev)
{
	DBG_PRINT("em86xx_eth_init\n");

	if (dev == NULL)
  		goto failed;

	rxdsc = (volatile struct em86xx_desc *)NON_CACHED((u32)eth_rxdsc);
	txdsc = (volatile struct em86xx_desc *)NON_CACHED((u32)eth_txdsc);
	rxbuf = (volatile unsigned char *)NON_CACHED((u32)eth_rxbuf);
	txbuf = (volatile unsigned char *)NON_CACHED((u32)eth_txbuf);

	/* Ethernet device initialization */
#ifdef BOOTLOADER
        /* Fill in the fields of the device structure with ethernet-generic values.*/
//      dev->type               = ARPHRD_ETHER;
        dev->mtu                = ETH_DATA_LEN; /* eth_mtu */
        dev->addr_len           = ETH_ALEN;
        memset(dev->broadcast, 0xff, ETH_ALEN);
        dev->hard_header_len = ETH_HLEN;
#else
	ether_setup(dev);
#endif

	if (em86xx_set_mac(dev))
  		goto failed;

	/* Hook up with handlers */
#ifdef BOOTLOADER
        dev->send_packet         = em86xx_eth_tx;
        dev->open                = em86xx_eth_open;
        dev->close               = em86xx_eth_close;
#else
	dev->get_stats		 = em86xx_eth_stats;
	dev->hard_start_xmit	 = em86xx_eth_tx;
	dev->open		 = em86xx_eth_open;
	dev->stop		 = em86xx_eth_close;
#ifdef USE_HW_FILTERING
	dev->set_multicast_list  = em86xx_eth_set_multicast_list;
#endif
	dev->do_ioctl		 = em86xx_eth_ioctl;
	dev->set_mac_address	 = em86xx_eth_set_macaddr;

	dev->tx_queue_len = num_txdesc; 
//	dev->flags &= ~IFF_MULTICAST;
//	dev->flags |= IFF_DEBUG;
#endif
	return 0;

failed:
	return(-EIO);
}

static int em86xx_eth_reset_desc(struct net_device *dev, int *reset)
{
	EM86XX_ETH_PRIV *private;
	int i;

	if (dev == NULL)
  		return(-EIO);
	else
		private = (EM86XX_ETH_PRIV *)dev->priv;

	if (*reset)
  		memset(dev->priv, 0, sizeof(EM86XX_ETH_PRIV));

	*reset = 0;

	/* Clear all tx/rx buffer memory */
	memset((void *)(rxbuf), 0, num_rxdesc * R_BUF_SIZE);
	memset((void *)(txbuf), 0, num_txdesc * T_BUF_SIZE);

	/* Initialize the indices */
	private->last_rxidx = 0;
	private->next_txidx = 0;

#ifndef BOOTLOADER
	if (rx_skb_list != NULL) {
		for (i = 0; i < num_rxdesc; i++) {
			if (rx_skb_list[i] != NULL)
				dev_kfree_skb(rx_skb_list[i]);
		}
	}
	memset((void *)rx_skb_list, 0, sizeof(struct skb_buff *) * num_rxdesc);
#endif

	/* Intialize the descriptors */
	em86xx_eth_setup_desc();

	return 0;
}

/* Driver installation: this is where the thing starts */
#ifdef BOOTLOADER
int em86xx_eth_probe(struct net_device *dev)
#else
static int __init em86xx_eth_startup(void)
#endif
{
#ifndef BOOTLOADER
	struct net_device *dev 	= &em86xx_eth_dev;
#endif
	int err = 0, i;
	EM86XX_ETH_PRIV *private= NULL;
#ifndef STATIC_BUF_ALLOC
	unsigned long max_num_rxdsc, max_num_txdsc;
	unsigned long rxbuf_size, txbuf_size;
#endif

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
        extern int tango2_ethernet_enabled(void);
        if (tango2_ethernet_enabled() == 0) {
                MSG_PRINT(KERN_WARNING "Tango2 builtin ethernet is disabled.\n");
                return(0);
        }
#endif
	/* Turn off IRQ and stop receive/transmit */
	em86xx_write_reg(EM86XX_CR_REG, 0);
	em86xx_write_reg(EM86XX_IER_REG, 0);
	em86xx_write_reg(EM86XX_SR_REG, em86xx_read_reg(EM86XX_SR_REG));

	if (dev_count != 0) {
		err = -EIO;
  		goto failed;
	} 

#ifndef STATIC_BUF_ALLOC	
	/* Validating module parameters */
	max_num_rxdsc = max_num_txdsc = PAGE_SIZE / (2 * sizeof(struct em86xx_desc));
#ifdef BOOTLOADER
	if (((num_rxdesc < MIN_NUM_RDESC) || (num_rxdesc > max_num_rxdsc)) ||
	    ((num_txdesc < MIN_NUM_TDESC) || (num_txdesc > max_num_txdsc))) {
		err = -EIO;
		goto failed;
	} else if ((desc_page = NON_CACHED(malloc(PAGE_SIZE))) == 0) {
		err = -ENOMEM;
		goto failed;
	}
#else
	if (((num_rxdesc < MIN_NUM_RDESC) || (num_rxdesc > max_num_rxdsc)) ||
	    ((num_txdesc < MIN_NUM_TDESC) || (num_txdesc > max_num_txdsc))) {
		err = -EIO;
		goto failed;
	} else if ((desc_page = __get_free_page(GFP_KERNEL | GFP_DMA)) == 0) {
		err = -ENOMEM;
		goto failed;
	} else if ((rx_skb_list = (struct skb_buff **)kmalloc(sizeof(struct skb_buff *) 
				                 * num_rxdesc, GFP_KERNEL)) == NULL) {
		err = -ENOMEM;
		goto failed;
	}

	memset((void *)rx_skb_list, 0, sizeof(struct skb_buff *) * num_rxdesc);
#endif

	/* Split a page for both rx/tx descriptor */
	eth_rxdsc = (struct em86xx_desc *)desc_page;
	eth_txdsc = (struct em86xx_desc *)(desc_page + ((1 << PAGE_SHIFT) / 2));

	/* Calculate the size needed for tx/rx -- aligned by pages */
	rxbuf_size = num_rxdesc * R_BUF_SIZE;
	txbuf_size = num_txdesc * T_BUF_SIZE;
	if ((rxbuf_size & (PAGE_SIZE - 1)) != 0)
		rxbuf_size = (rxbuf_size & PAGE_MASK) + PAGE_SIZE;
	if ((txbuf_size & (PAGE_SIZE - 1)) != 0)
		txbuf_size = (txbuf_size & PAGE_MASK) + PAGE_SIZE;
	DBG_PRINT("txbuf_size = 0x%08lx, rxbuf_size = 0x%08lx\n", txbuf_size, rxbuf_size);

#ifndef BOOTLOADER
	/* Calculate the order needed for tx/rx */
	for (rxbuf_order = 0; rxbuf_order < MAX_ORDER; rxbuf_order++) {
		if ((PAGE_SIZE * (1 << rxbuf_order)) >= rxbuf_size)
			break;
	}

	for (txbuf_order = 0; txbuf_order < MAX_ORDER; txbuf_order++) {
		if ((PAGE_SIZE * (1 << txbuf_order)) >= txbuf_size)
			break;
	}

	if ((rxbuf_order >= MAX_ORDER) || (txbuf_order >= MAX_ORDER)) {
		err = -ENOMEM;
		goto failed;
	}

	DBG_PRINT("txbuf_order = 0x%08lx, rxbuf_order = 0x%08lx\n", txbuf_order, rxbuf_order);
#endif
		
#ifdef BOOTLOADER
	if ((rxbuf_pages = NON_CACHED(malloc(rxbuf_size))) == 0) {
		err = -ENOMEM;
		goto failed;
	} else if ((txbuf_pages = NON_CACHED(malloc(txbuf_size))) == 0) {
		err = -ENOMEM;
		goto failed;
	}
#else
	if ((rxbuf_pages = __get_free_pages(GFP_KERNEL | GFP_DMA, rxbuf_order)) == 0) {
		err = -ENOMEM;
		goto failed;
	} else if ((txbuf_pages = __get_free_pages(GFP_KERNEL | GFP_DMA, txbuf_order)) == 0) {
		err = -ENOMEM;
		goto failed;
	}
#endif

	eth_rxbuf  = (unsigned char *)rxbuf_pages;
	eth_txbuf  = (unsigned char *)txbuf_pages;
#endif

#ifndef BOOTLOADER
	SET_MODULE_OWNER(&em86xx_eth_dev);
#endif

	MSG_PRINT("Ethernet driver for EM86XX (v1.0)");
	dev->priv = NULL;
        dev->irq = IRQ_ETHERNET;

	/* Set up default mac address and others */
	max_link_loop 	= DEF_LINK_LOOP;
	ld_count 	= 1;
	reset_flag 	= 1;

	/* Get a device name: normally it'll be eth0 */
#ifdef BOOTLOADER
        memcpy( dev->name, DRIVER, sizeof(DRIVER));

        /* Register ISR */
        em86xx_request_irq(IRQ_ETHERNET, em86xx_eth_intr_handler, dev);
#else
	if ((err = dev_alloc_name(&em86xx_eth_dev, "eth%d")) < 0)
  		goto failed;

	/* Register ISR */
	if ((err = request_irq(IRQ_ETHERNET, em86xx_eth_intr_handler, 0, "ethernet", 
      	 	 &em86xx_eth_dev)) != 0) {
  		err = -EIO;
  		goto failed;
	} else
#endif 
  		MSG_PRINT(" on %s (IRQ: %d)", dev->name, IRQ_ETHERNET);


	if (em86xx_get_macaddr(dev->dev_addr)) {
  		err = -EIO;
  		goto failed;
	}

	MSG_PRINT("\n(MAC %02x", dev->dev_addr[0]);
	for (i = 1; i < 6; i++)
	  	MSG_PRINT(":%02x", dev->dev_addr[i]);
	MSG_PRINT(", tx_desc/rx_desc = %ld/%ld), ", num_txdesc, num_rxdesc);
	if (phy_address < 0)
		MSG_PRINT("PHY probing enabled\n");
	else
		MSG_PRINT("PHY address: %d\n", phy_address);

	/* Allocate memory for private data */
#ifdef BOOTLOADER
        dev->priv = private = (EM86XX_ETH_PRIV *)malloc(sizeof(EM86XX_ETH_PRIV));//, GFP_KERNEL);
#else
	dev->priv = private = (EM86XX_ETH_PRIV *)kmalloc(sizeof(EM86XX_ETH_PRIV), GFP_KERNEL);
#endif
	if (dev->priv == NULL) {
		err = -ENOMEM;
		goto failed;
	}

	memset(dev->priv, 0, sizeof(EM86XX_ETH_PRIV));

	/* Point to driver initialization routine and register the device with kernel */
	dev->init = em86xx_eth_init;
#ifdef BOOTLOADER
        em86xx_eth_init(dev);
        dev->state = NETDEV_DOWN;
#else
	if (register_netdev(dev) != 0) {
  		err = -EIO;
  		goto failed;
	}
#endif

	MSG_PRINT("%s: driver installation completed.\n", dev->name);
	dev_count++;
	return 0;

failed:
  	if (dev->priv != NULL) {
#ifdef BOOTLOADER
                free(dev->priv);
#else
    		kfree(dev->priv);
#endif
    		dev->priv = NULL;
  	}
#ifndef STATIC_BUF_ALLOC
#ifdef BOOTLOADER
	if (desc_page   != 0)
		free((void *)CACHED(desc_page));
	if (rxbuf_pages != 0)
		free((void *)CACHED(rxbuf_pages));
	if (txbuf_pages != 0)
		free((void *)CACHED(txbuf_pages));
#else
	if (desc_page   != 0)
		free_page(desc_page);
	if (rxbuf_pages != 0)
		free_pages(rxbuf_pages, rxbuf_order);
	if (txbuf_pages != 0)
		free_pages(txbuf_pages, txbuf_order);
	if (rx_skb_list != NULL);
		kfree(rx_skb_list);
#endif
#endif
  	MSG_PRINT("%s: driver installation failed.\n", dev->name);
  	return(err);
}

#ifndef BOOTLOADER
/* Uninstallation of drive */
static void __exit em86xx_eth_shutdown(void)
{
#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
        extern int tango2_ethernet_enabled(void);
        if (tango2_ethernet_enabled() == 0) 
                return;
#endif
	if (dev_count != 0) {
		/* Turn off IRQ and stop receive/transmit */
  		em86xx_write_reg(EM86XX_CR_REG, 0);
  		em86xx_write_reg(EM86XX_IER_REG, 0);

  		/* Unregister the device and ISR */
  		free_irq(IRQ_ETHERNET, &em86xx_eth_dev);
		unregister_netdev(&em86xx_eth_dev);

		/* Set desc base address registers to 0 */
		em86xx_write_reg(EM86XX_RLBAR_REG, 0);
		em86xx_write_reg(EM86XX_TLBAR_REG, 0);
#ifndef STATIC_BUF_ALLOC
#ifdef BOOTLOADER
		if (desc_page   != 0)
			free((void *)CACHED(desc_page));
		if (rxbuf_pages != 0)
			free((void *)CACHED(rxbuf_pages));
		if (txbuf_pages != 0)
			free((void *)CACHED(txbuf_pages));
#else
		if (desc_page   != 0)
			free_page(desc_page);
		if (rxbuf_pages != 0)
			free_pages(rxbuf_pages, rxbuf_order);
		if (txbuf_pages != 0)
			free_pages(txbuf_pages, txbuf_order);
		if (rx_skb_list != NULL) {
			int i;
			for (i = 0; i < num_rxdesc; i++) {
				if (rx_skb_list[i] != NULL)
					dev_kfree_skb(rx_skb_list[i]);
			}
			kfree(rx_skb_list);
		}
#endif
#endif
		/* Free up memory */
		if (em86xx_eth_dev.priv != NULL) {
			kfree(em86xx_eth_dev.priv);
			em86xx_eth_dev.priv = NULL;
  		}
	} 
}

/* Register startup/shutdown routines */
module_init(em86xx_eth_startup);
module_exit(em86xx_eth_shutdown);
#endif

